#  _MultiplicativeScatterCorrection.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import builtins
from abc import ABC, abstractmethod
from typing import Optional, List, Tuple

from ..core import utils, real
from ..core.matrix import Matrix, helper
from ._AbstractTransformation import AbstractTransformation


class MultiplicativeScatterCorrection(AbstractTransformation):
    def __init__(self):
        super().__init__()
        self.prefilter: Optional[AbstractTransformation] = None
        self.correction: AbstractMultiplicativeScatterCorrection = RangeBased()
        self.average: Optional[Matrix] = None

    @staticmethod
    def create_configuration_matrix(spectra: List[Matrix]) -> Matrix:
        if len(spectra) == 0:
            raise ValueError("Can't create configuration matrix for 0 spectra")

        wave_numbers = spectra[0].get_column(0)
        amplitudes = [spectrum.get_column(1) for spectrum in spectra]

        return helper.multi_concat(1, wave_numbers, *amplitudes)

    def configure(self, data: Matrix):
        wave_numbers = data.get_column(0)
        amplitudes = data.get_sub_matrix((0, data.num_rows()), (1, data.num_columns()))

        self.average = wave_numbers.concat_along_columns(amplitudes.mean(1))
        self.configured = True

    def do_transform(self, data: Matrix) -> Matrix:
        return self.correction.correct(self.average, data)

    def do_inverse_transform(self, data: Matrix) -> Matrix:
        raise NotImplementedError('Cannot perform inverse transform of Multiplicative Scatter Correction')


class AbstractMultiplicativeScatterCorrection(ABC):
    def __init__(self):
        self.prefilter: Optional[AbstractTransformation] = None

    @abstractmethod
    def correct(self, average: Matrix, data: Matrix) -> Matrix:
        pass


class RangeBased(AbstractMultiplicativeScatterCorrection):
    def __init__(self):
        super().__init__()

        self.ranges: List[Tuple[real, real]] = []

    def correct(self, average: Matrix, data: Matrix) -> Matrix:
        result = data.copy()

        if self.prefilter is not None:
            filtered = self.prefilter.transform(data)
        else:
            filtered = data

        for range in self.ranges:
            x = []
            y = []
            wave = []
            for i in builtins.range(average.num_rows()):
                if self.range_contains(range, filtered.get(i, 0)):
                    wave.append(filtered.get(i, 0))
                    y.append(filtered.get(i, 1))
                    x.append(average.get(i, 1))

            inter, slope = utils.linear_regression(x, y)

            for i in builtins.range(result.num_rows()):
                if self.range_contains(range, result.get(i, 0)):
                    result.set(i, 1, (result.get(i, 1) - inter) / slope)

        return result

    @staticmethod
    def range_contains(range: Tuple[real, real], value: real) -> bool:
        lower, upper = range

        return lower <= value <= upper



#  _AbstractMultiResponsePLS.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from abc import abstractmethod
from typing import List, Optional

from ...core import real, PreprocessingType
from ...core.matrix import Matrix, helper
from ...transformation import AbstractTransformation, Center, Standardize
from ._AbstractPLS import AbstractPLS


class AbstractMultiResponsePLS(AbstractPLS):
    """
    Ancestor for PLS algorithms that work on multiple response variables.
    """
    def __init__(self):
        super().__init__()
        self.class_mean: Optional[List[real]] = None
        self.class_std_dev: Optional[List[real]] = None
        self.trans_predictors: Optional[AbstractTransformation] = None
        self.trans_response: Optional[AbstractTransformation] = None

    def reset(self):
        """
        Resets the member variables.
        """
        super().reset()

        self.class_mean = None
        self.class_std_dev = None
        self.trans_predictors = None
        self.trans_response = None

    @abstractmethod
    def get_min_columns_response(self) -> int:
        """
        Returns the minimum number of columns ther response matrix has to have.

        :return:    The minimum.
        """
        pass

    @abstractmethod
    def get_max_columns_response(self) -> int:
        """
        Returns the maximum number of columns the response matrix has to have.

        :return:    The maximum, -1 for unlimited.
        """
        pass

    def check(self, predictors: Optional[Matrix], response: Optional[Matrix]) -> Optional[str]:
        """
        Hook method for checking the data before training.

        :param predictors:  The input data.
        :param response:    The dependent variable(s).
        :return:            None if successful, otherwise error message.
        """
        result = super().check(predictors, response)

        if result is None:
            if response.num_columns() < self.get_min_columns_response():
                result = 'Algorithm requires at least ' + \
                         str(self.get_min_columns_response()) + \
                         ' response columns, found: ' + \
                         str(response.num_columns())
            elif self.get_max_columns_response() != -1 and response.num_columns() > self.get_max_columns_response():
                result = 'Algorithm can handle at most ' + \
                         str(self.get_max_columns_response()) + \
                         ' response columns, found: ' + \
                         str(response.num_columns())

        return result

    def do_initialize(self, predictors: Matrix, response: Matrix) -> Optional[str]:
        """
        Initializes using the provided data.

        :param predictors:  The input data.
        :param response:    The dependent variable(s).
        :return:            None if successful, otherwise error message.
        """
        self.class_mean = [real(0)] * response.num_columns()
        self.class_std_dev = [real(0)] * response.num_columns()
        for i in range(response.num_columns()):
            if self.preprocessing_type is PreprocessingType.CENTER:
                self.class_mean[i] = helper.mean(response, 0)
                self.class_std_dev[i] = real(1)
                self.trans_predictors = Center()
                self.trans_response = Center()
            elif self.preprocessing_type is PreprocessingType.STANDARDIZE:
                self.class_mean[i] = helper.mean(response, 0)
                self.class_std_dev[i] = helper.stdev(response, 0)
                self.trans_predictors = Standardize()
                self.trans_response = Standardize()
            elif self.preprocessing_type is PreprocessingType.NONE:
                self.class_mean[i] = real(0)
                self.class_std_dev[i] = real(1)
                self.trans_predictors = None
                self.trans_response = None
            else:
                raise RuntimeError('Unhandled preprocessing type: ' + str(self.preprocessing_type))

        if self.trans_predictors is not None:
            self.trans_predictors.configure(predictors)
            predictors = self.trans_predictors.transform(predictors)
        if self.trans_response is not None:
            self.trans_response.configure(response)
            response = self.trans_response.transform(response)

        result = self.do_perform_initialization(predictors, response)

        return result

    def do_predict(self, predictors: Matrix) -> Matrix:
        """
        Performs predictions on the data.

        :param predictors:  The input data.
        :return:            The transformed data and the predictions.
        """
        result = self.do_perform_predictions(predictors)
        if self.trans_response is not None:
            for i in range(result.num_rows()):
                for j in range(result.num_columns()):
                    result.set(i, j, result.get(i, j) * self.class_std_dev[j] + self.class_mean[j])

        return result



#  _SavitzkyGolayTest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional

from wai.test.decorators import RegressionTest, ExceptionTest
from wai.ma.core.matrix import Matrix
from wai.ma.transformation import SavitzkyGolay

from ..test import Tags
from ._AbstractTransformationTest import AbstractTransformationTest


class SavitzkyGolayTest(AbstractTransformationTest):
    @classmethod
    def subject_type(cls):
        return SavitzkyGolay

    @RegressionTest
    def coefficients_3_3_2_0(self, subject: SavitzkyGolay, bolts: Matrix):
        self.coefficient_helper(subject, 3, 3, 2, 0)
        return self.standard_regression(subject, bolts)

    @RegressionTest
    def coefficients_3_3_2_1(self, subject: SavitzkyGolay, bolts: Matrix):
        self.coefficient_helper(subject, 3, 3, 2, 1)
        return self.standard_regression(subject, bolts)

    @RegressionTest
    def coefficients_3_3_2_2(self, subject: SavitzkyGolay, bolts: Matrix):
        self.coefficient_helper(subject, 3, 3, 2, 2)
        return self.standard_regression(subject, bolts)

    @RegressionTest
    def coefficients_0_3_2_1(self, subject: SavitzkyGolay, bolts: Matrix):
        self.coefficient_helper(subject, 0, 3, 2, 1)
        return self.standard_regression(subject, bolts)

    @RegressionTest
    def coefficients_3_0_2_1(self, subject: SavitzkyGolay, bolts: Matrix):
        self.coefficient_helper(subject, 3, 0, 2, 1)
        return self.standard_regression(subject, bolts)

    @ExceptionTest(NotImplementedError)
    def check_inv_transform_eq_input(self, subject: SavitzkyGolay, bolts: Matrix):
        """
        Savitzky-Golay doesn't have an inverse transform.
        """
        super().check_inv_transform_eq_input()

    def standard_regression(self, subject: SavitzkyGolay, *resources: Matrix):
        bolts, = resources

        transform: Matrix = subject.transform(bolts)

        return {
            Tags.TRANSFORM: transform,
            "coefficients": subject.coefficients,
            "serialised": subject.get_serialised_state()
        }

    @staticmethod
    def coefficient_helper(subject: SavitzkyGolay,
                           num_points_left: Optional[int] = None,
                           num_points_right: Optional[int] = None,
                           polynomial_order: Optional[int] = None,
                           derivative_order: Optional[int] = None):
        if polynomial_order is not None:
            subject.polynomial_order = polynomial_order
        if derivative_order is not None:
            subject.derivative_order = derivative_order
        if num_points_left is not None:
            subject.num_points_left = num_points_left
        if num_points_right is not None:
            subject.num_points_right = num_points_right



#  _DIPLSTest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from wai.test.decorators import RegressionTest, Test

from wai.ma.algorithm.pls import DIPLS
from wai.ma.core.matrix import Matrix
from wai.ma.core.matrix.factory import randn_like

from ._AbstractPLSTest import AbstractPLSTest
from ...test import Tags


class DIPLSTest(AbstractPLSTest):
    @classmethod
    def subject_type(cls):
        return DIPLS

    @RegressionTest
    def lambda_01(self, subject: DIPLS, bolts: Matrix, bolts_response: Matrix):
        subject.lambda_ = 0.01
        return self.standard_regression(subject, bolts, bolts_response)

    def standard_regression(self, subject: DIPLS, *resources: Matrix):
        x_source_domain, y_source_domain = resources
        x_target_domain: Matrix = x_source_domain.add(randn_like(x_source_domain, 0, 1, 2))
        y_target_domain: Matrix = y_source_domain.add(randn_like(y_source_domain, 1, 1, 2))
        x_target_domain_unlabeled: Matrix = x_source_domain.add(randn_like(x_source_domain, 100, 1, 2))

        result = {}

        # Initialise supervised
        subject.initialize_supervised(x_source_domain, x_target_domain, y_source_domain, y_target_domain)
        result.update(self.add_default_pls_matrices(subject, x_target_domain, Tags.SUPERVISED))
        subject.reset()

        # Initialise unsupervised
        subject.initialize_unsupervised(x_source_domain, x_target_domain, y_source_domain)
        result.update(self.add_default_pls_matrices(subject, x_target_domain, Tags.UNSUPERVISED))
        subject.reset()

        # Initialise semisupervised
        subject.initialize_semisupervised(x_source_domain, x_target_domain, x_target_domain_unlabeled,
                                          y_source_domain, y_target_domain)
        result.update(self.add_default_pls_matrices(subject, x_target_domain, Tags.SEMISUPERVISED))

        return result

    @Test
    def check_transformed_num_components(self, subject: DIPLS, bolts: Matrix, bolts_response: Matrix):
        X, Y = bolts, bolts_response
        X_2: Matrix = X.add(randn_like(X, 0, 0, 2))

        for i in range(1, 5):
            subject.num_components = i
            subject.initialize_unsupervised(X, X_2, Y)
            transform: Matrix = subject.transform(X)
            self.assertEqual(i, transform.num_columns())

            # Reset
            subject = self.instantiate_subject()



#  _SupervisedFilter.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from abc import abstractmethod
from typing import Optional

from .matrix import Matrix
from ._Filter import Filter


class SupervisedFilter(Filter):
    """
    Interface for filters which are supervised, and therefore require
    training before their transform method can be used.
    """
    @abstractmethod
    def initialize(self, predictors: Matrix, response: Matrix) -> Optional[str]:
        """
        Initialises using the provided data.
        
        :param predictors:  The input data.
        :param response:    The dependent variable(s).
        :return:            None if successful, otherwise error message.
        """
        pass



#  _MatrixSerialiser.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import IO, Optional, List, Tuple

from wai.test.serialisation import RegressionSerialiser
from wai.ma.core.matrix import Matrix, helper

from ._epsilon import EPSILON


class MatrixSerialiser(RegressionSerialiser[Matrix]):
    @classmethod
    def binary(cls) -> bool:
        return False

    @classmethod
    def extension(cls) -> str:
        return "csv"

    @classmethod
    def serialise(cls, result: Matrix, file: IO[str]):
        helper.write(result, file, False, ',', 10, False)

    @classmethod
    def deserialise(cls, file: IO[str]) -> Matrix:
        return helper.read(file, False, ",")

    @classmethod
    def compare(cls, result: Matrix, reference: Matrix) -> Optional[str]:
        # Check for shape
        if not reference.same_shape_as(result):
            return 'Shapes of the expected and actual matrices do not match.\n' + \
                   'Expected shape: ' + reference.shape_string() + '\n' + \
                   'Actual shape: ' + result.shape_string()

        # Check for values
        diff: Matrix = reference.sub(result).abs()
        which = diff.which(lambda v: v > EPSILON)
        if len(which) > 0:
            return 'Values in the expected and actual matrices do not match.\n' + \
                   'Absolute differences:\n' + \
                   cls.indices_to_string(which, diff)

    @classmethod
    def indices_to_string(cls, which: List[Tuple[int, int]], diff: Matrix) -> str:
        string = '(<row>, <column>): <absolute difference>\n'
        string += '\n'.join(('({i}, {j}): {diff}'.format_map({'i': i, 'j': j, 'diff': diff.get(i, j)})
                            for i, j in which))
        return string



#  _MatrixTest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Tuple

from wai.ma.core.matrix import Matrix, factory
from wai.test.decorators import RegressionTest, Test

from ...test import TestDataset, Tags, AbstractMatrixAlgorithmTest


class MatrixTest(AbstractMatrixAlgorithmTest):
    @classmethod
    def subject_type(cls):
        return Matrix

    @classmethod
    def instantiate_subject(cls) -> Matrix:
        return factory.randn(7, 7, seed=2)

    @classmethod
    def get_datasets(cls) -> Tuple[TestDataset]:
        return TestDataset.BOLTS,

    @RegressionTest
    def transpose(self, subject: Matrix, bolts: Matrix):
        return self.standard_regression(subject.transpose())

    @RegressionTest
    def get_eigenvectors(self, subject: Matrix, bolts: Matrix):
        return self.standard_regression(subject.get_eigenvectors())

    @RegressionTest
    def get_eigenvalues(self, subject: Matrix, bolts: Matrix):
        return self.standard_regression(subject.get_eigenvalues())

    @RegressionTest
    def svd_S(self, subject: Matrix, bolts: Matrix):
        return self.standard_regression(subject.svd_S())

    @RegressionTest
    def svd_U(self, subject: Matrix, bolts: Matrix):
        return self.standard_regression(subject.svd_U())

    @RegressionTest
    def svd_V(self, subject: Matrix, bolts: Matrix):
        return self.standard_regression(subject.svd_V())

    @RegressionTest
    def norm2(self, subject: Matrix, bolts: Matrix):
        return self.standard_regression(bolts.norm2())

    @RegressionTest
    def norm1(self, subject: Matrix, bolts: Matrix):
        return self.standard_regression(bolts.norm1())

    @Test
    def column_is_sub_matrix(self, subject: Matrix, bolts: Matrix):
        for i in range(subject.num_columns()):
            column = subject.get_column(i)
            sub_matrix = subject.get_sub_matrix((0, subject.num_rows()),
                                                (i, i + 1))
            self.assertEqual(column, sub_matrix)

    def standard_regression(self, subject: Matrix, *resources):
        # Only regression testing the final configuration of the matrix
        return {
            Tags.MATRIX: subject
        }



#  _PLS1.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional, List

from ._AbstractSingleResponsePLS import AbstractSingleResponsePLS
from ...core import real
from ...core.matrix import Matrix, factory, helper


class PLS1(AbstractSingleResponsePLS):
    def __init__(self):
        super().__init__()
        self.r_hat: Optional[Matrix] = None  # The regression vector "r-hat"
        self.P: Optional[Matrix] = None  # The P matrix
        self.W: Optional[Matrix] = None  # The W matrix
        self.b_hat: Optional[Matrix] = None  # The b-hat vector

    def reset(self):
        """
        Resets the member variables.
        """
        super().reset()

        self.r_hat = None
        self.P = None
        self.W = None
        self.b_hat = None

    def get_matrix_names(self) -> List[str]:
        """
        Returns the names of all the available matrices.

        :return:    The names of the matrices.
        """
        return ['RegVector',
                'P',
                'W',
                'b_hat']

    def get_matrix(self, name: str) -> Optional[Matrix]:
        """
        Returns the matrix with the specified name.

        :param name:    The name of the matrix.
        :return:        The matrix, None if not available.
        """
        if name == 'RegVector':
            return self.r_hat
        elif name == 'P':
            return self.P
        elif name == 'W':
            return self.W
        elif name == 'b_hat':
            return self.b_hat
        else:
            return None

    def has_loadings(self) -> bool:
        """
        Whether the algorithm supports the return of loadings.

        :return:    True if supported.
        """
        return True

    def get_loadings(self) -> Optional[Matrix]:
        """
        Returns the loadings, if available.

        :return:     The loadings, None if not available.
        """
        return self.get_matrix('P')

    def do_perform_initialization(self, predictors: Matrix, response: Matrix) -> Optional[str]:
        """
        Initializes using the provided data.

        :param predictors:  The input data.
        :param response:    The dependent variable(s).
        :return:            None if successful, otherwise error message.
        """
        X_k: Matrix = predictors
        y: Matrix = response

        # Init
        W: Matrix = factory.zeros(predictors.num_columns(), self.num_components)
        P: Matrix = factory.zeros(predictors.num_columns(), self.num_components)
        T: Matrix = factory.zeros(predictors.num_rows(), self.num_components)
        b_hat: Matrix = factory.zeros(self.num_components, 1)

        for k in range(self.num_components):
            # 1. step: wj
            w_k: Matrix = self.calculate_weights(X_k, y)
            W.set_column(k, w_k)

            # 2. step: tj
            t_k: Matrix = X_k.mul(w_k)
            T.set_column(k, t_k)

            # 3. step: ^bj
            tdott: real = t_k.vector_dot(t_k)
            b_k: real = real(t_k.vector_dot(y) / tdott)
            b_hat.set(k, 0, b_k)

            # 4. step: pj
            p_k: Matrix = X_k.transpose().mul(t_k).div(tdott)
            P.set_column(k, p_k)

            # 5. step: Xk+1 (deflating y is not necessary)
            X_k = X_k.sub(t_k.mul(p_k.transpose()))

        # W*(P^T*W)^-1
        tmp: Matrix = W.mul(((P.transpose()).mul(W)).inverse())

        # factor = W*(P^T*W)^-1 * b_hat
        self.r_hat = tmp.mul(b_hat)

        # Save matrices
        self.P = P
        self.W = W
        self.b_hat = b_hat

        return None

    def calculate_weights(self, x_k: Matrix, y: Matrix) -> Matrix:
        """
        Calculate the weight w_k in the PLS iterations.

        :param x_k:     X matrix at step k.
        :param y:       y matrix.
        :return:        Weights at step k.
        """
        return x_k.transpose().mul(y).normalized()

    def do_transform(self, predictors: Matrix) -> Matrix:
        """
        Transforms the data.

        :param predictors:  The input data.
        :return:            The transformed data and the predictions.
        """
        result = factory.zeros(predictors.num_rows(), self.num_components)

        for i in range(predictors.num_rows()):
            # Work on each row
            x: Matrix = helper.row_as_vector(predictors, i)
            X: Matrix = factory.zeros(1, self.num_components)
            T: Matrix = factory.zeros(1, self.num_components)

            for j in range(self.num_components):
                X.set_column(j, x)
                # 1. step: tj = xj * wj
                t: Matrix = x.mul(self.W.get_column(j))
                T.set_column(j, t)
                # 2. step:xj+1 = xj - tj*pj^T (tj is 1x1 matrix!)
                x = x.sub(self.P.get_column(j).transpose().mul(t.as_real()))

            result.set_row(i, T)

        return result

    def can_predict(self) -> bool:
        """
        Returns whether the algorithm can make predictions.

        :return:    True if can make predictions.
        """
        return True

    def do_perform_predictions(self, predictors: Matrix) -> Matrix:
        """
        Performs predictions on the data.

        :param predictors:  The input data.
        :return:            The transformed data and the predictions.
        """
        result = factory.zeros(predictors.num_rows(), 1)

        for i in range(predictors.num_rows()):
            x: Matrix = helper.row_as_vector(predictors, i)
            X: Matrix = factory.zeros(1, self.num_components)
            T: Matrix = factory.zeros(1, self.num_components)

            for j in range(self.num_components):
                X.set_column(j, x)
                # 1. step: tj = xj * wj
                t: Matrix = x.mul(self.W.get_column(j))
                T.set_column(j, t)
                # 2. step: xj+1 = xj - tj*pj^T (tj is 1x1 matrix!)
                x = x.sub(self.P.get_column(j).transpose().mul(t.as_real()))

            result.set(i, 0, T.mul(self.b_hat).as_real())

        return result



#  _Standardize.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, IO

from ..core import real, Serialisable
from ..core.matrix import Matrix
from ..core.matrix.helper import column_stdevs, column_means
from ._AbstractTransformation import AbstractTransformation


class Standardize(AbstractTransformation, Serialisable):
    def __init__(self):
        super().__init__()
        self.means: Optional[Matrix] = None
        self.std_devs: Optional[Matrix] = None

    def reset(self):
        super().reset()
        self.means = None
        self.std_devs = None

    def configure(self, data: Matrix):
        self.means = column_means(data)
        self.std_devs = column_stdevs(data)
        # Make sure we don't do a divide by zero
        for i in range(self.std_devs.num_rows()):
            if self.std_devs.get(i, 0) == real(0):
                self.std_devs.set(i, 0, 1)
        self.configured = True

    def do_transform(self, data: Matrix) -> Matrix:
        result = data.copy()
        result.sub_by_vector_modify(self.means)
        result.div_by_vector_modify(self.std_devs)
        return result

    def do_inverse_transform(self, data: Matrix) -> Matrix:
        result = data.copy()
        result.mul_by_vector_modify(self.std_devs)
        result.add_by_vector_modify(self.means)
        return result

    def get_means(self) -> Matrix:
        return self.means

    def get_std_devs(self) -> Matrix:
        return self.std_devs

    def serialise_state(self, stream: IO[bytes]):
        # Can't serialise our state until we've been configured
        if not self.configured:
            raise RuntimeError("Can't serialise state of unconfigured Standardize")

        # Serialise out our mean and standard deviation coefficient matrices
        self.means.serialise_state(stream)
        self.std_devs.serialise_state(stream)



#  _FFT.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from enum import Enum

import numpy as np
from wai.common import switch, case, default

from ..core.matrix import Matrix
from ._AbstractTransformation import AbstractTransformation


class OutputMode(Enum):
    """
    Enumeration of the different types of output the FFT transformation
    can produce.
    """
    REAL_COMPONENT_ONLY = 1
    IMAG_COMPONENT_ONLY = 2
    REAL_AND_IMAG_PAIRS = 3
    AMPLITUDE_ONLY = 4
    PHASE_ANGLE_ONLY = 8
    AMPL_ANGLE_PAIRS = 12


class FFT(AbstractTransformation):
    def __init__(self):
        super().__init__()

        # Options
        self.orthonormalisation: bool = False
        self.output_mode: OutputMode = OutputMode.AMPLITUDE_ONLY

    def __setattr__(self, key, value):
        # Can set output mode by int as well
        if key == "output_mode" and isinstance(value, int):
            value = OutputMode(value)

        # Otherwise just as normal
        super().__setattr__(key, value)

    def configure(self, data: Matrix):
        # No config needed
        self.configured = True

    def do_transform(self, data: Matrix) -> Matrix:
        samples: np.ndarray = data.data

        fourier: np.ndarray = self.fft(samples)

        internal_formatted: np.ndarray = self.internal_format(fourier)

        output_formatted: np.ndarray = self.format_output(internal_formatted)

        return Matrix(output_formatted)

    def do_inverse_transform(self, data: Matrix) -> Matrix:
        output_formatted: np.ndarray = data.data

        internal_formatted: np.ndarray = self.inverse_format_output(output_formatted)

        fourier: np.ndarray = self.internal_unformat(internal_formatted)

        samples: np.ndarray = self.ifft(fourier)

        return Matrix(samples)

    def fft(self, array: np.ndarray) -> np.ndarray:
        """
        Gets the FFT of the given array.

        :param array:           The array to FFT.
        :return:                The FFT of the array.
        """
        return np.fft.fftshift(
            np.fft.fft(
                array,
                axis=1,
                norm="ortho" if self.orthonormalisation else None
            )
        )

    def ifft(self, array: np.ndarray) -> np.ndarray:
        """
        Gets the inverse FFT of the given array.

        :param array:           The array to inverse FFT.
        :return:                The inverse FFT of the array.
        """
        return np.fft.ifft(
            np.fft.ifftshift(array),
            axis=1,
            norm="ortho" if self.orthonormalisation else None
        )

    @staticmethod
    def internal_format(fourier: np.ndarray) -> np.ndarray:
        """
        Formats the Fourier transform (complex) array into a real
        array consisting of the real components, optionally paired with
        the necessary imaginary components of the complex numbers.

        :param fourier:     The result of the FFT.
        :return:            The internally-formatted array.
        """
        # Get the number of columns in the array
        num_columns: int = fourier.shape[1]

        # Whether we need the imaginary component of the Nyquist frequency
        need_nyquist: bool = num_columns % 2 == 0

        # Initialise the formatted columns with the real part of the zero-frequency term
        columns = [np.real(fourier[:, num_columns // 2])]

        # Add the real and imaginary parts of the positive frequencies
        for i in range(num_columns // 2 + 1, num_columns):
            column = fourier[:, i]
            columns.append(np.real(column))
            columns.append(np.imag(column))

        # Add the real part of the Nyquist frequency if needed
        if need_nyquist:
            columns.append(np.real(fourier[:, 0]))

        # Return the stacked columns
        return np.stack(columns, axis=1)

    @staticmethod
    def internal_unformat(internal_formatted: np.ndarray) -> np.ndarray:
        """
        Inverse of internal format. Restores FFT result from internally-formatted
        array.

        :param internal_formatted:  The internally-formatted array.
        :return:                    The FFT.
        """
        # Get the number of columns
        num_columns: int = internal_formatted.shape[1]

        # Work out if we have a Nyquist frequency component
        has_nyquist: bool = num_columns % 2 == 0

        # Create lists of the real and imaginary columns,
        # initialised with the zero-frequency term
        real_columns = [internal_formatted[:, 0]]
        imag_columns = [np.zeros_like(real_columns[0])]

        # Add the formatted columns to their respective lists
        for i in range(1, num_columns):
            respective_list = real_columns if i % 2 == 1 else imag_columns
            respective_list.append(internal_formatted[:, i])

        # Add the (zero) imaginary part of the Nyquist term,
        # if it's present
        if has_nyquist:
            imag_columns.append(np.zeros_like(real_columns[0]))

        # Recreate the negative frequencies
        real_columns = [column for column in reversed(real_columns[1:])] + real_columns
        imag_columns = [-1 * column for column in reversed(imag_columns[1:])] + imag_columns

        # If we has a Nyquist frequency, we now have two,
        # so discard the end one
        if has_nyquist:
            real_columns = real_columns[:-1]
            imag_columns = imag_columns[:-1]

        # Recompose the real and imaginary parts
        real = np.stack(real_columns, axis=1)
        imag = 1j * np.stack(imag_columns, axis=1)

        # Return the complex sum
        return real + imag

    def format_output(self, array: np.ndarray) -> np.ndarray:
        """
        Formats the (complex) output array based on the given mode. Inverse
        function of inverse_format_output.

        :param array:   The complex array.
        :return:        The formatted array
        """
        with switch(self.output_mode):
            if case(OutputMode.REAL_COMPONENT_ONLY):
                return format_real_component_only(array)
            if case(OutputMode.IMAG_COMPONENT_ONLY):
                return format_imag_component_only(array)
            if case(OutputMode.REAL_AND_IMAG_PAIRS):
                return format_real_and_imag_pairs(array)
            if case(OutputMode.AMPLITUDE_ONLY):
                return format_amplitude_only(array)
            if case(OutputMode.PHASE_ANGLE_ONLY):
                return format_phase_angle_only(array)
            if case(OutputMode.AMPL_ANGLE_PAIRS):
                return format_ampl_angle_pairs(array)

    def inverse_format_output(self, array: np.ndarray) -> np.ndarray:
        """
        Removes the formatting on the given array for the given output mode.
        Inverse function of format_output.

        :param array:   The formatted array.
        :return:        The unformatted array.
        """
        with switch(self.output_mode):
            if case(OutputMode.REAL_AND_IMAG_PAIRS):
                return inverse_format_real_and_imag_pairs(array)
            if case(OutputMode.AMPL_ANGLE_PAIRS):
                return inverse_format_ampl_angle_pairs(array)
            if default():
                raise ValueError("Inverse transform only available for modes pair modes, not " + self.output_mode.name)


def format_real_component_only(array: np.ndarray) -> np.ndarray:
    columns = [array[:, 0]]

    for i in range(1, array.shape[1], 2):
        columns.append(array[:, i])

    return np.stack(columns, axis=1)


def format_imag_component_only(array: np.ndarray) -> np.ndarray:
    columns = [np.zeros_like(array[:, 0])]

    for i in range(2, array.shape[1], 2):
        columns.append(array[:, i])

    if array.shape[1] % 2 == 0:
        columns.append(np.zeros_like(array[:, 0]))

    return np.stack(columns, axis=1)


def format_real_and_imag_pairs(array: np.ndarray) -> np.ndarray:
    return array


def format_amplitude_only(array: np.ndarray) -> np.ndarray:
    columns = [array[:, 0] + 0j]

    for i in range(2, array.shape[1], 2):
        columns.append(array[:, i - 1] + 1j * array[:, i])

    if array.shape[1] % 2 == 0:
        columns.append(array[:, -1] + 0j)

    return np.abs(np.stack(columns, axis=1))


def format_phase_angle_only(array: np.ndarray) -> np.ndarray:
    columns = [array[:, 0] + 0j]

    for i in range(2, array.shape[1], 2):
        columns.append(array[:, i - 1] + 1j * array[:, i])

    if array.shape[1] % 2 == 0:
        columns.append(array[:, -1] + 0j)

    return np.angle(np.stack(columns, axis=1))


def format_ampl_angle_pairs(array: np.ndarray) -> np.ndarray:
    columns = [array[:, 0]]

    for i in range(2, array.shape[1], 2):
        compl = array[:, i - 1] + 1j * array[:, i]
        columns.append(np.abs(compl))
        columns.append(np.angle(compl))

    if array.shape[1] % 2 == 0:
        columns.append(array[:, -1])

    return np.stack(columns, axis=1)


def inverse_format_real_and_imag_pairs(array: np.ndarray) -> np.ndarray:
    return array


def inverse_format_ampl_angle_pairs(array: np.ndarray) -> np.ndarray:
    columns = [array[:, 0]]

    for i in range(2, array.shape[1], 2):
        ampl = array[:, i - 1]
        angle = array[:, i]
        compl = ampl * (np.cos(angle) + 1j * np.sin(angle))
        columns.append(np.real(compl))
        columns.append(np.imag(compl))

    if array.shape[1] % 2 == 0:
        columns.append(array[:, -1])

    return np.stack(columns, axis=1)



#  _SIMPLS.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional, List, IO

from ._AbstractSingleResponsePLS import AbstractSingleResponsePLS
from ...core import real, utils, Serialisable
from ...core.matrix import Matrix, factory
from ...core.utils import sqrt


class SIMPLS(AbstractSingleResponsePLS, Serialisable):
    def __init__(self):
        super().__init__()
        self.num_coefficients: int = 0  # The number of coefficients in W to keep (0 keep all)
        self.W: Optional[Matrix] = None  # The W matrix
        self.B: Optional[Matrix] = None  # The B matrix (used for prediction)
        self.Q: Optional[Matrix] = None  # Q matrix to regress T (XW) on y

    def initialize(self, predictors: Optional[Matrix] = None, response: Optional[Matrix] = None) -> Optional[str]:
        """
        Initialises the members.
        """
        if predictors is None and response is None:
            super().initialize()
            self.num_coefficients = 0
        else:
            return super().initialize(predictors, response)

    def reset(self):
        """
        Resets the member variables.
        """
        super().reset()
        self.B = None
        self.W = None

    @staticmethod
    def validate_num_coefficients(value: int) -> bool:
        return True

    def get_matrix_names(self) -> List[str]:
        """
        Returns all available matrices.

        :return:    The names of the matrices.
        """
        return ['W', 'B', 'Q']

    def get_matrix(self, name: str) -> Optional[Matrix]:
        """
        Returns the matrix with the specified name.

        :param name:    The name of the matrix.
        :return:        The matrix, None if not available.
        """
        if name == 'W':
            return self.W
        elif name == 'B':
            return self.B
        elif name == 'Q':
            return self.Q
        else:
            return None

    def has_loadings(self) -> bool:
        """
        Whether the algorithm supports the return of loadings.

        :return:    True if supported.
        """
        return True

    def get_loadings(self) -> Optional[Matrix]:
        """
        Returns the loadings, if available.

        :return:    The loadings, None if not available.
        """
        return self.W

    def slim(self, in_: Matrix):
        """
        Zeroes the coefficients of the W matrix beyond the specified number of
        coefficients.

        :param in_: The matrix to process in-place.
        """
        B: List[List[real]] = in_.to_raw_copy_2D()

        for i in range(in_.num_columns()):
            l: Matrix = in_.get_column(i)
            ld: List[real] = l.to_raw_copy_1D()
            ld = [abs(x) for x in ld]
            srt: List[int] = utils.sort(ld)
            index: int = srt[max(len(srt) - 1 - self.num_coefficients, 0)]

            val: real = ld[index]
            for c in range(in_.num_rows()):
                if abs(B[c][i]) < val:
                    in_.set(c, i, 0)

    def do_perform_initialization(self, predictors: Matrix, response: Matrix) -> Optional[str]:
        """
        Initializes using the provided data.

        :param predictors:  The input data.
        :param response:    The dependent variable(s).
        :return:            None if successful, otherwise error message.
        """
        X_trans: Matrix = predictors.transpose()
        A: Matrix = X_trans.mul(response)
        M: Matrix = X_trans.mul(predictors)
        C: Matrix = factory.eye(predictors.num_columns(), predictors.num_columns())
        W: Matrix = factory.zeros(predictors.num_columns(), self.num_components)
        P: Matrix = factory.zeros(predictors.num_columns(), self.num_components)
        Q: Matrix = factory.zeros(1, self.num_components)

        for h in range(self.num_components):
            # // 1. qh as dominant EigenVector of Ah'*Ah
            A_trans: Matrix = A.transpose()
            q: Matrix = A_trans.mul(A).get_dominant_eigenvector()

            # 2. wh=Ah*qh, ch=wh'*Mh*wh, wh=wh/sqrt(ch), store wh in W as column
            w: Matrix = A.mul(q)
            c: Matrix = w.transpose().mul(M).mul(w)
            w = w.mul(1.0 / sqrt(c.as_real()))
            W.set_column(h, w)

            # 3. ph=Mh*wh, store ph in P as column
            p: Matrix = M.mul(w)
            p_trans: Matrix = p.transpose()
            P.set_column(h, p)

            # 4. qh=Ah'*wh, store qh in Q as column
            q: Matrix = A_trans.mul(w)
            Q.set_column(h, q)

            # 5. vh=Ch*ph, vh=vh/||vh||
            v: Matrix = C.mul(p)
            v = v.normalized()
            v_trans: Matrix = v.transpose()

            # 6. Ch+1=Ch-vh*vh', Mh+1=Mh-ph*ph'
            C = C.sub(v.mul(v_trans))
            M = M.sub(p.mul(p_trans))

            # 7. Ah+1=ChAh (actually Ch+1)
            A = C.mul(A)

        # Finish
        if self.num_coefficients > 0:
            self.slim(W)
        self.W = W
        self.B = W.mul(Q.transpose())
        self.Q = Q

        return None

    def do_transform(self, predictors: Matrix) -> Matrix:
        """
        Transforms the data.

        :param predictors:  The input data.
        :return:            The transformed data.
        """
        return predictors.mul(self.W)

    def can_predict(self) -> bool:
        """
        Returns whether the algorithm can make predictions.

        :return:    True if can make predictions.
        """
        return True

    def do_perform_predictions(self, predictors: Matrix) -> Matrix:
        """
        Performs predictions on the data.

        :param predictors:  The input data.
        :return:            The predictions.
        """
        return predictors.mul(self.B)

    def serialise_state(self, stream: IO[bytes]):
        # Can't serialise our state until we've been initialised
        if not self.initialised:
            raise RuntimeError("Can't serialise state of uninitialised SIMPLS")

        # Serialise out our W matrix
        self.W.serialise_state(stream)



#  _PCA.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional, List

from ..transformation import Center, Standardize, AbstractTransformation
from ..core import real, utils, Filter
from ..core.matrix import Matrix, factory
from ._AbstractAlgorithm import AbstractAlgorithm


class PCA(AbstractAlgorithm, Filter):
    def __init__(self):
        super().__init__()
        self.variance: real = real(0.95)
        self.max_columns: int = -1
        self.center: bool = False
        self.loadings: Optional[Matrix] = None
        self.scores: Optional[Matrix] = None
        self.correlation: Optional[Matrix] = None
        self.eigenvectors: Optional[Matrix] = None
        self.eigenvalues: Optional[Matrix] = None
        self.sorted_eigens: List[int] = []
        self.sum_of_eigenvalues: real = real(0)
        self.num_cols: int = 0
        self.num_rows: int = 0
        self.keep_cols: List[int] = []
        self.train: Optional[Matrix] = None
        self.transformation: Optional[AbstractTransformation] = None

    def initialize(self):
        super().initialize()
        self.variance = real(0.95)
        self.max_columns = -1
        self.center = False

    def reset(self):
        super().reset()
        self.loadings = None
        self.scores = None
        self.correlation = None
        self.eigenvectors = None
        self.eigenvalues = None
        self.sorted_eigens = []
        self.sum_of_eigenvalues = real(0)
        self.transformation = None

    @staticmethod
    def validate_variance(value: real):
        return real(0) < value < real(1)

    @staticmethod
    def validate_max_columns(value: int):
        return value == -1 or value > 0

    @staticmethod
    def validate_center(value: bool):
        return True

    def fill_correlation(self):
        """
        Fills the covariance matrix.
        """
        self.correlation = factory.zeros(self.num_cols, self.num_cols)

        for i in range(self.num_cols):
            for j in range(i, self.num_cols):
                cov = real(0)
                for n in range(self.train.num_rows()):
                    cov += self.train.get(n, i) * self.train.get(n, j)

                cov /= self.train.num_rows() - 1
                self.correlation.set(i, j, cov)
                self.correlation.set(j, i, cov)

    def remove_columns(self, data: Matrix) -> Matrix:
        """
        Removes the columns according to self.keep_cols.

        :param data:    The data to trim.
        :return:        The trimmed data.
        """
        if len(self.keep_cols) != data.num_columns():
            rows = []
            for j in range(data.num_rows()):
                rows.append(j)
            data = data.get_sub_matrix(rows, self.keep_cols)

        return data

    def configure(self, instances: Matrix):
        """
        Initialises the filter with the given input data.

        :param instances:   The data to process.
        """
        self.train = instances.copy()

        # Delete any attributes with only one distinct value or are all missing
        self.keep_cols = []
        for j in range(self.train.num_columns()):
            distinct = set()
            for i in range(self.train.num_rows()):
                distinct.add(self.train.get(i, j))
                if len(distinct) > 1:
                    break
            if len(distinct) > 1:
                self.keep_cols.append(j)
        self.train = self.remove_columns(self.train)

        # Transform data
        if self.center:
            self.transformation = Center()
        else:
            self.transformation = Standardize()
        self.train = self.transformation.transform(self.train)

        self.num_rows = self.train.num_rows()
        self.num_cols = self.train.num_columns()

        self.fill_correlation()

        # Get eigenvectors/values
        corr = self.correlation.copy()
        self.eigenvectors = corr.get_eigenvectors().copy()
        self.eigenvalues = corr.get_eigenvalues().copy()

        # Any eigenvalues less than 0 are not worth anything -- change to 0
        self.eigenvalues = self.eigenvalues.apply_elementwise(lambda v: real(0) if v < 0 else v)

        self.sorted_eigens = utils.sort(self.eigenvalues.to_raw_copy_1D())
        self.sum_of_eigenvalues = sum(self.eigenvalues.to_raw_copy_1D())

        self.train = None

    def do_transform(self, data: Matrix) -> Matrix:
        """
        Transform a matrix.

        :param data:    The original data to transform.
        :return:        The transformed data.
        """
        num_cols = self.max_columns if self.max_columns > 0 else self.num_cols
        if self.max_columns > 0:
            num_cols_lower_bound = self.num_cols - self.max_columns
        else:
            num_cols_lower_bound = 0
        if num_cols_lower_bound < 0:
            num_cols_lower_bound = 0

        data = self.remove_columns(data)
        data = self.transformation.transform(data)
        values = [[]] * data.num_rows()
        num_cols_act = 0
        for n in range(data.num_rows()):
            new_vals = [real(0)] * num_cols

            cumulative = 0
            cols = 0
            for i in reversed(range(num_cols_lower_bound, self.num_cols)):
                cols += 1
                val = real(0)
                for j in range(self.num_cols):
                    val += self.eigenvectors.get_row(j).get(0, self.sorted_eigens[i]) * data.get(n, j)

                new_vals[self.num_cols - i - 1] = val
                cumulative += self.eigenvalues.get_from_vector(self.sorted_eigens[i])
                if (cumulative / self.sum_of_eigenvalues) >= self.variance:
                    break

            num_cols_act = max(num_cols_act, cols)
            values[n] = new_vals

        if self.debug:
            self.logger.info('num_cols_act: ' + str(num_cols_act))

        # Generate matrix based on actual number of retained columns
        result = factory.zeros(data.num_rows(), num_cols_act)
        for n in range(len(values)):
            for i in range(min(len(values[n]), num_cols_act)):
                result.set(n, i, values[n][i])

        return result

    def get_coefficients(self) -> Optional[List[List[real]]]:
        if self.eigenvalues is None:
            return None

        if self.max_columns > 0:
            num_cols_lower_bound = self.num_cols - self.max_columns
        else:
            num_cols_lower_bound = 0
        if num_cols_lower_bound < 0:
            num_cols_lower_bound = 0

        # All of the coefficients for a single principle component
        result = []
        cumulative = real(0)

        # Loop through each principle component
        for i in reversed(range(num_cols_lower_bound, self.num_cols)):
            one_pc = []

            for j in range(self.num_cols):
                coeff_value = self.eigenvectors.get(j, self.sorted_eigens[i])
                one_pc.append(coeff_value)

            result.append(one_pc)
            cumulative += self.eigenvalues.get_from_vector(self.sorted_eigens[i])

            if (cumulative / self.sum_of_eigenvalues) >= self.variance:
                break

        return result

    def extract_loadings(self) -> Matrix:
        """
        Create a matrix to output from the coefficients 2D list.

        :return:    Matrix containing the components.
        """
        coeff = self.get_coefficients()
        result = factory.zeros(self.num_cols, len(coeff) + 1)

        # Add the index column
        for n in range(self.num_cols):
            result.set(n, result.num_columns() - 1, n + 1)

        # Each list is a single column
        for i in range(len(coeff)):
            for n in range(self.num_cols):
                # Column was kept earlier
                value = real(0)
                if n in self.keep_cols:
                    index = self.keep_cols.index(n)
                    if index < len(coeff[i]):
                        value = coeff[i][index]

                result.set(n, i, value)

        return result

    def transform(self, data: Matrix) -> Matrix:
        """
        Transforms the data.

        :param data:    The data to transform.
        :return:        The transformed data.
        """
        self.reset()
        self.configure(data)
        result = self.do_transform(data)
        self.scores = result
        self.loadings = self.extract_loadings()

        return result

    def to_string(self) -> str:
        """
        For outputting some information about the algorithm.

        :return:    The information.
        """
        result =    self.__class__.__name__ + '\n'
        result +=   self.__class__.__name__.replace('.', '=') + '\n\n'
        result +=   'Debug      : ' + str(self.debug) + '\n'
        result +=   'Variance   : ' + str(self.variance) + '\n'
        result +=   'Max columns: ' + str(self.max_columns) + '\n'
        result +=   'Center     : ' + str(self.center) + '\n'

        return result








#  _YGradientGLSW.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import List, Optional

from ._GLSW import GLSW
from ...core import real
from ...core.error import MatrixAlgorithmsError
from ...core.matrix import Matrix, factory
from ...transformation import AbstractTransformation


class YGradientGLSW(GLSW):
    def get_covariance_matrix(self, X: Matrix, y: Matrix) -> Matrix:
        y_vals: List[real] = y.to_raw_copy_1D()
        sorted_increasing_row_indices: List[int] = [i for i in range(len(y_vals))]
        sorted_increasing_row_indices.sort(key=lambda o: y_vals[o])

        # Sort increasing
        all_cols: List[int] = [i for i in range(X.num_columns())]
        X_sorted: Matrix = X.get_sub_matrix(sorted_increasing_row_indices, all_cols)
        y_sorted: Matrix = y.get_sub_matrix(sorted_increasing_row_indices, [0])

        sav_golay: SavitzkyGolayFilter = SavitzkyGolayFilter()

        # Apply 5-Point first derivative Savitzky-Golay filter
        X_smoothed: Matrix = sav_golay.transform(X_sorted)
        y_smoothed: Matrix = sav_golay.transform(y_sorted)

        y_smoothed_mean: real = y_smoothed.mean(-1).as_real()
        syd: real = y_smoothed.sub(y_smoothed_mean).pow_elementwise(2).sum(-1).div(y_smoothed.num_rows() - 1).sqrt().as_real()

        # Reweighting matrix
        W: Matrix = factory.zeros(y.num_rows(), y.num_rows())
        for i in range(y_smoothed.num_rows()):
            ydi: real = y_smoothed.get(i, 0)
            W.set(i, i, pow(2.0, -1 * ydi / syd))

        # Covariance matrix
        C: Matrix = X_smoothed.transpose().mul(W.mul(W)).mul(X_smoothed)
        return C

    def check(self, x1: Matrix, x2: Matrix) -> Optional[str]:
        """
        Hook method for checking the data before training.

        :param x1:  First sample set.
        :param x2:  Second sample set.
        :return:    None if successful,
                    otherwise error message.
        """
        if x1 is None:
            return 'No x1 matrix provided!'
        if x2 is None:
            return 'No x2 matrix provided!'
        if x1.num_rows() != x2.num_rows():
            return 'Predictors and response must have the same number of rows!'
        return None


class SavitzkyGolayFilter(AbstractTransformation):
    coef: List[real] = [real(x) for x in [2.0 / 10.0,
                                          1.0 / 10.0,
                                          0.0,
                                          -1.0 / 10.0,
                                          -2.0 / 10.0]]

    def configure(self, data: Matrix):
        self.configured = True

    def do_transform(self, data: Matrix) -> Matrix:
        mat_extended: Matrix = self.extend_matrix(data)
        result: Matrix = factory.zeros_like(mat_extended)

        # Smooth all rows
        for i in range(2, mat_extended.num_rows() - 2):
            row_smoothed: Matrix = self.smooth_row(i, mat_extended)
            result.set_row(i, row_smoothed)

        orig_sized_matrix: Matrix = self.shrink_matrix(result)
        return orig_sized_matrix

    def shrink_matrix(self, result: Matrix) -> Matrix:
        """
        Shrink matrix to the original size. Removes first and last 2 rows.

        :param result:  Input matrix.
        :return:        Shrunk matrix.
        """
        return result.get_sub_matrix((2, result.num_rows() - 2), (0, result.num_columns()))

    def extend_matrix(self, data: Matrix) -> Matrix:
        """
        Extend the matrix by 2 rows in the beginning and 2 rows
        at the end (copy first and last elements).

        :param data:    Input matrix.
        :return:        Extended matrix.
        """
        # Extend the matrix by 2 rows at the beginning and 2 rows at the end.
        first_row: Matrix = data.get_row(0)
        last_row: Matrix = data.get_row(data.num_rows() - 1)
        return first_row.concat_along_rows(first_row)\
                        .concat_along_rows(data)\
                        .concat_along_rows(last_row)\
                        .concat_along_rows(last_row)

    def do_inverse_transform(self, data: Matrix) -> Matrix:
        raise MatrixAlgorithmsError('Inverse transform of Savitzky-Golay is not available.')

    def smooth_row(self, i: int, matrix: Matrix) -> Matrix:
        """
        Apply first five-point first gradient Savitzky-Golay smoothing to the i-th
        row of the given matrix.

        :param i:       Row index.
        :param matrix:  Input matrix.
        :return:        Smoothed row.
        """
        res: Matrix = factory.zeros(1, matrix.num_columns())

        window_size: int = len(SavitzkyGolayFilter.coef)
        for m in range(window_size):
            coef_idx: int = (window_size - 1) - m
            row_idx: int = i - (m - 2)
            row: Matrix = matrix.get_row(row_idx)
            row_scaled: Matrix = row.mul(SavitzkyGolayFilter.coef[coef_idx])
            res = res.add(row_scaled)

        return res



#  _AbstractPLS.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from abc import abstractmethod
from typing import List, Optional

from ...core import SupervisedFilter
from ...algorithm import AbstractAlgorithm
from ...core import PreprocessingType
from ...core.matrix import Matrix


class AbstractPLS(AbstractAlgorithm, SupervisedFilter):
    def __init__(self):
        super().__init__()
        self.preprocessing_type: PreprocessingType = PreprocessingType.NONE  # The preprocessing type to perform
        self.num_components: int = 5  # The maximum number of components to generate

    def initialize(self, predictors: Optional[Matrix] = None, response: Optional[Matrix] = None) -> Optional[str]:
        if predictors is None and response is None:
            """
            Initializes the members.
            """
            super().initialize()
            self.num_components = 5
            self.preprocessing_type = PreprocessingType.NONE
        else:
            """
            Initialises using the provided data.

            :param predictors:  The input data.
            :param response:    The dependent variable(s).
            :return:            None if successful, otherwise error message.
            """
            # Always work on copies
            predictors = predictors.copy()
            response = response.copy()

            self.reset()

            result = self.check(predictors, response)

            if result is None:
                result = self.do_initialize(predictors, response)
                self.initialised = result is None

            return result

    @staticmethod
    def validate_preprocessing_type(value: PreprocessingType) -> bool:
        return True

    @staticmethod
    def validate_num_components(value: int) -> bool:
        return True

    @abstractmethod
    def get_matrix_names(self) -> List[str]:
        """
        Returns all the available matrices.

        :return:    The names of the matrices.
        """
        pass

    @abstractmethod
    def get_matrix(self, name: str) -> Optional[Matrix]:
        """
        Returns the matrix with the specified name.

        :param name:    The name of the matrix.
        :return:        The matrix, None if not available.
        """
        pass

    @abstractmethod
    def has_loadings(self) -> bool:
        """
        Whether the algorithm supports the return of loadings.

        :return:    True if supported.
        """
        pass

    @abstractmethod
    def get_loadings(self) -> Optional[Matrix]:
        """
        Returns the loadings.

        :return:    The loadings, None if not available.
        """
        pass

    def check(self, predictors: Optional[Matrix], response: Optional[Matrix]) -> Optional[str]:
        """
        Hook method for checking the data before training.

        :param predictors:  The input data.
        :param response:    The dependent variable(s).
        :return:            None if successful, otherwise error message.
        """
        if predictors is None:
            return 'No predictors matrix provided!'
        if response is None:
            return 'No response matrix provided!'
        return None

    @abstractmethod
    def do_initialize(self, predictors: Matrix, response: Matrix) -> Optional[str]:
        """
        Trains using the provided data.

        :param predictors:  The input data.
        :param response:    The dependent variable(s).
        :return:            None if successful, otherwise error message.
        """
        pass

    @abstractmethod
    def can_predict(self) -> bool:
        """
        Returns whether the algorithm can make predictions.

        :return:    True if can make predictions.
        """
        pass

    @abstractmethod
    def do_predict(self, predictors: Matrix) -> Matrix:
        """
        Performs predictions on the data.

        :param predictors:  The input data.
        :return:            The predictions.
        """
        pass

    def predict(self, predictors: Matrix) -> Matrix:
        """
        Performs predictions on the data.

        :param predictors:  The input data.
        :return:            The predictions.
        """
        if not self.is_initialised():
            raise RuntimeError('Algorithm has not been initialised!')

        return self.do_predict(predictors)

    @abstractmethod
    def do_transform(self, predictors: Matrix) -> Matrix:
        """
        Transforms the data.

        :param predictors:  The input data.
        :return:            The transformed data.
        """
        pass

    def transform(self, predictors: Matrix) -> Matrix:
        """
        Transforms the data.

        :param predictors:  The input data.
        :return:            The transformed data.
        """
        if not self.is_initialised():
            raise RuntimeError('Algorithm has not been initialised!')

        return self.do_transform(predictors)

    @abstractmethod
    def do_perform_initialization(self, predictors: Matrix, response: Matrix) -> Optional[str]:
        """
        Initialises using the provided data.

        :param predictors:  The input data.
        :param response:    The dependent variable(s).
        :return:            None if successful, otherwise error message.
        """
        pass

    @abstractmethod
    def do_perform_predictions(self, predictors: Matrix) -> Matrix:
        """
        Performs predictions on the data.

        :param predictors:  The input data.
        :return:            The transformed data and the predictions.
        """
        pass

    def to_string(self) -> str:
        """
        For outputting some information about the algorithm.

        :return:    The information.
        """
        result = self.__class__.__name__ + '\n'
        result += self.__class__.__name__.replace('.', '=') + '\n\n'
        result += 'Debug        : ' + str(self.debug) + '\n'
        result += '# components : ' + str(self.num_components) + '\n'
        result += 'Preprocessing: ' + str(self.preprocessing_type) + '\n'

        return result



#  _AbstractMatrixAlgorithmTest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from abc import ABC, abstractmethod
from typing import Tuple, Dict, Type, Any

from wai.test.decorators import RegressionTest

from wai.ma.core import real
from wai.ma.core.matrix import Matrix
from wai.test import AbstractTest
from wai.test.serialisation import RegressionSerialiser

from ._MatrixSerialiser import MatrixSerialiser
from ._RealSerialiser import RealSerialiser
from ._TestDataset import TestDataset


class AbstractMatrixAlgorithmTest(AbstractTest, ABC):
    @classmethod
    @abstractmethod
    def get_datasets(cls) -> Tuple[TestDataset, ...]:
        """
        Get the input datasets used for the algorithm tests.

        :return:    Paths to input data.
        """
        pass

    @abstractmethod
    def standard_regression(self, subject, *resources) -> Dict[str, Any]:
        """
        Defines the standard way to regression-test the subject with
        the given resources.

        :param subject:     The subject being tested.
        :param resources:   The resources being used to test the subject.
        :return:            The regression map.
        """
        pass

    @classmethod
    def common_resources(cls) -> Tuple[Matrix, ...]:
        return tuple(dataset.load() for dataset in cls.get_datasets())

    @classmethod
    def common_serialisers(cls) -> Dict[Type, Type[RegressionSerialiser]]:
        return {
            Matrix: MatrixSerialiser,
            real: RealSerialiser
        }

    @RegressionTest
    def default_setup(self, subject, *resources):
        """
        Applies the standard regression method to the default subject.

        :param subject:     The default subject.
        :param resources:   Any resources required for the regression test.
        :return:            A regression map.
        """
        return self.standard_regression(subject, *resources)



#  _SparsePLSTest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from wai.test.decorators import RegressionTest

from wai.ma.algorithm.pls import SparsePLS
from wai.ma.core.matrix import Matrix

from ._AbstractPLSTest import AbstractPLSTest


class SparsePLSTest(AbstractPLSTest):
    @classmethod
    def subject_type(cls):
        return SparsePLS

    @classmethod
    def instantiate_subject(cls) -> SparsePLS:
        spls: SparsePLS = SparsePLS()
        spls.num_components = 2
        return spls

    @RegressionTest
    def lambda_0(self, subject: SparsePLS, *resources: Matrix):
        subject.lambda_ = 0
        return self.standard_regression(subject, *resources)

    @RegressionTest
    def lambda_05(self, subject: SparsePLS, *resources: Matrix):
        subject.lambda_ = 0.001
        return self.standard_regression(subject, *resources)



#  _YGradientGLSWTest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Tuple

from test.test import TestDataset
from wai.ma.algorithm.glsw import YGradientGLSW
from wai.ma.core.matrix import Matrix

from ._GLSWTest import GLSWTest


class YGradientGLSWTest(GLSWTest):
    @classmethod
    def subject_type(cls):
        return YGradientGLSW

    def standard_regression(self, subject: YGradientGLSW, *resources: Matrix):
        # Get inputs: Simulate second instrument as x1 with noise
        X, y = resources

        # Init GLSW
        subject.initialize(X, y)

        # Add regressions
        return self.add_glsw_regressions(subject, X)

    @classmethod
    def get_datasets(cls) -> Tuple[TestDataset, TestDataset]:
        return TestDataset.BOLTS, TestDataset.BOLTS_RESPONSE



#  helper.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import sys
from functools import partial
from numbers import Number
from typing import List, IO, Union

import numpy as np

from ...core.utils import real_to_string_fixed
from ...meta import print_stack_trace
from . import factory
from ._Matrix import Matrix
from .._types import real


def delete_col(data: Matrix, col: int) -> Matrix:
    return delete_cols(data, [col])


def delete_cols(data: Matrix, cols: List[int]) -> Matrix:
    cols = [col for col in range(data.num_columns()) if col not in cols]
    return data.get_sub_matrix((0, data.num_rows()), cols)


def row_as_vector(data: Matrix, row: int) -> Matrix:
    return data.get_row(row)


def mean(mat: Matrix, index: int, column: bool = True) -> real:
    if column:
        return column_mean(mat, index)
    else:
        return row_mean(mat, index)


def stdev(mat: Matrix, index: int, column: bool = True) -> real:
    if column:
        return column_stdev(mat, index)
    else:
        return row_stdev(mat, index)


def column_as_vector(m: Matrix, column_index: int) -> Matrix:
    return m.get_column(column_index)


def euclidean_distance(x: Matrix, y: Matrix, squared: bool) -> Matrix:
    xx: Matrix = row_norms(x, True)
    yy: Matrix = row_norms(y, True)

    distances: Matrix = x.mul(y.transpose())
    distances = distances.mul(-2)
    distances = distances.add(xx)
    distances = distances.add(yy)

    if x == y:
        fill_diagonal(distances, 0)

    if squared:
        return distances
    else:
        return distances.sqrt()


def row_norms(x: Matrix, squared: bool) -> Matrix:
    # TODO
    raise NotImplementedError


def equal(m1: Matrix, m2: Matrix, epsilon: real = real(0)) -> bool:
    if not m1.same_shape_as(m2):
        return False

    for row in range(m1.num_rows()):
        for col in range(m1.num_columns()):
            if np.abs(m1.get(row, col) - m2.get(row, col)) > epsilon:
                return False

    return True


def read(filename: Union[str, IO[str]], header: bool, separator: str) -> Matrix:
    """
    Reads the matrix from the given CSV file.

    :param filename:    The file to read from.
    :param header:      True if the file contains a header (gets skipped).
    :param separator:   The column separator used.
    :return:            The matrix.
    """
    if isinstance(filename, str):
        with open(filename, 'r') as file:
            lines = [line for line in file]
    else:
        lines = [line for line in filename]

    if len(lines) == 0:
        raise RuntimeError('No rows in file: ' + filename)

    if header:
        lines = lines[1:]
    if len(lines) == 0:
        raise RuntimeError('No data rows in file: ' + filename)

    sep = separator
    cells = lines[0].split(sep)
    result = factory.zeros(len(lines), len(cells))
    for i in range(len(lines)):
        cells = lines[i].split(sep)
        for j in range(min(len(cells), result.num_columns())):
            try:
                result.set(i, j, real(cells[j]))
            except Exception:
                print('Failed to parse row=' + str(i + 1 if header else i) + ' col=' + str(j) + ': ' + str(cells[j]),
                      file=sys.stderr)
                print_stack_trace()

    return result


def to_lines(data: Matrix, header: bool, separator: str, num_dec: int, scientific: bool = False) -> List[str]:
    result: List[str] = []

    if header:
        result.append(separator.join(['col' + str(j + 1) for j in range(data.num_columns())]))

    data_to_string(data, separator, num_dec, result, scientific)

    return result


def data_to_string(data: Matrix, separator: str, num_dec: int, result: List[str], scientific: bool):
    if scientific:
        formatter = partial(np.format_float_scientific,
                            precision=num_dec,
                            unique=True,
                            trim='0',
                            exp_digits=1)
    elif num_dec == -1:
        formatter = str
    else:
        formatter = partial(real_to_string_fixed, after_decimal_point=num_dec)

    def get_formatted(row, column):
        return formatter(data.get(row, column))

    def line_gen(row):
        return separator.join(get_formatted(row, column) for column in range(data.num_columns()))

    for row in range(data.num_rows()):
        result.append(line_gen(row))


def write(data: Matrix, filename: Union[str, IO[str]], header: bool, separator: str, num_dec: int, scientific: bool = False):
    if isinstance(filename, str):
        with open(filename, 'w') as file:
            file.writelines((line + '\n' for line in to_lines(data, header, separator, num_dec, scientific)))
    else:
        filename.writelines((line + '\n' for line in to_lines(data, header, separator, num_dec, scientific)))


def to_string(data: Matrix, header: bool = True, separator: str = '\t', num_dec: int = 6) -> str:
    return '\n'.join(to_lines(data, header, separator, num_dec))


def dim(m: Matrix) -> str:
    return str(m.num_rows()) + ' x ' + str(m.num_columns())


def inverse(m: Matrix) -> Matrix:
    return m.inverse()


def row_mean(mat: Matrix, index: int) -> real:
    return row_means(mat).get(index, 0)


def row_means(mat: Matrix) -> Matrix:
    result = Matrix(np.mean(mat.data, 1, real))
    result = result.transpose()
    return result


def column_mean(mat: Matrix, index: int) -> real:
    return column_means(mat).get(0, index)


def column_means(mat: Matrix) -> Matrix:
    return Matrix(np.mean(mat.data, 0, real))


def row_stdev(mat: Matrix, index: int) -> real:
    return row_stdevs(mat).get(index, 0)


def row_stdevs(mat: Matrix) -> Matrix:
    stdevs = Matrix(np.std(mat.data, 1, real, ddof=1))
    stdevs = stdevs.transpose()
    return stdevs


def column_stdev(mat: Matrix, index: int) -> real:
    return column_stdevs(mat).get(0, index)


def column_stdevs(mat: Matrix) -> Matrix:
    return Matrix(np.std(mat.data, 0, real, ddof=1))


def fill_diagonal(mat: Matrix, value: Number):
    i = 0
    while i < mat.num_rows() and i < mat.num_columns():
        mat.data[i][i] = real(value)
        i += 1


def solve(A: Matrix, b: Matrix) -> Matrix:
    return Matrix(np.linalg.solve(A.data, b.data))


def multi_concat(axis: int, *matrices: Matrix) -> Matrix:
    return Matrix(np.concatenate([matrix.data for matrix in matrices], axis))



#  _Serialisable.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import struct
from abc import ABC, abstractmethod
from io import BytesIO
from typing import IO, Union

# Whether to use little-endianness for serialisation
# (False means big-endian)
LITTLE_ENDIAN: bool = True


class Serialisable(ABC):
    """
    Mixin class to add serialisation to a class.
    """
    @abstractmethod
    def serialise_state(self, stream: IO[bytes]):
        """
        Serialises the state of the object to binary.

        :param stream:  The stream to serialise to.
        """
        pass

    def save_state(self, filename: str):
        """
        Saves the state of this algorithm to file.

        :param filename:    The name of the file to save to.
        """
        # Make sure the directory for the file exists
        path = os.path.dirname(filename)
        if not os.path.exists(path):
            os.makedirs(path, exist_ok=True)

        # Save the file
        with open(filename, 'wb') as file:
            self.serialise_state(file)

    def get_serialised_state(self) -> bytes:
        """
        Gets the serialised state of this object.
        """
        stream = BytesIO()

        self.serialise_state(stream)

        return stream.read()

    @staticmethod
    def serialise_to_bytes(value: Union[int, float, str]) -> bytes:
        """
        Helper method to serialise common types to bytes.

        :param value:   The value to serialise.
        :return:        The bytes representation of the value.
        """
        if isinstance(value, int):
            b = value.to_bytes(4, "little" if LITTLE_ENDIAN else "big", signed=True)
            return b
        elif isinstance(value, float):
            b = struct.pack("<d" if LITTLE_ENDIAN else ">d", value)
            return b
        elif isinstance(value, str):
            b = value.encode(encoding="utf-8")
            b = Serialisable.serialise_to_bytes(len(b)) + b
            return b
        else:
            raise TypeError("serialise_to_bytes only supports int, float and str types")

    @staticmethod
    def serialise_big_int(value: int) -> bytes:
        """
        Serialises an integer to binary. Can handle values that fall
        outside the 32-bit range of serialise_to_bytes.

        :param value:   The integer to serialise.
        :return:        The bytes representation of the value.
        """
        # Get the number of bits required to represent this value.
        bit_length: int = value.bit_length()

        # Get the number of bytes required to represent this value
        byte_length: int = (bit_length // 8) + (1 if bit_length % 8 != 0 else 0)

        # Get the binary representation of the integer
        b = value.to_bytes(byte_length, "little" if LITTLE_ENDIAN else "big", signed=True)

        # Prepend it with the binary representation of the bit length
        b = Serialisable.serialise_to_bytes(byte_length) + b

        return b



#  _CCAFilter.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional

from ..core import SupervisedFilter
from ._AbstractAlgorithm import AbstractAlgorithm
from ..core import real
from ..core.matrix import Matrix, factory
from ..transformation import Center


class CCAFilter(AbstractAlgorithm, SupervisedFilter):
    def __init__(self):
        super().__init__()
        self.lambda_X: real = real(1e-2)
        self.lambda_Y: real = real(1e-2)
        self.kcca: int = 1
        self.center_X: Optional[Center] = None
        self.center_Y: Optional[Center] = None
        self.proj_X: Optional[Matrix] = None
        self.proj_Y: Optional[Matrix] = None

    @staticmethod
    def validate_kcca(value: int):
        return value > 0

    @staticmethod
    def validate_lambda_X(value: real):
        return True

    @staticmethod
    def validate_lambda_Y(value: real):
        return True

    def to_string(self) -> str:
        return 'Canonical Correlation Analysis Filter (CCARegression)'

    def reset(self):
        super().reset()
        self.proj_X = None
        self.proj_Y = None

    def initialize(self, x: Optional[Matrix] = None, y: Optional[Matrix] = None) -> Optional[str]:
        if x is None or y is None:
            super().initialize()
            self.kcca = 1
            self.lambda_X = real(1e-2)
            self.lambda_Y = real(1e-2)
            self.center_X = Center()
            self.center_Y = Center()
        else:
            x = x.copy()
            y = y.copy()

            self.reset()

            result = self.check(x, y)

            if result is None:
                result = self.do_initialize(x, y)
                self.initialised = (result is None)

            return result

    def do_initialize(self, X: Matrix, Y: Matrix) -> Optional[str]:
        super().initialize()

        num_features: int = X.num_columns()
        num_targets: int = Y.num_columns()

        # Check if dimension kcca is valid
        if self.kcca > min(num_features, num_targets):
            return 'Projection dimension must be <= min(X.num_columns(), Y.num_columns()).'

        # Center input
        X = self.center_X.transform(X)
        Y = self.center_Y.transform(Y)

        # Regularisation matrices
        lambda_IX: Matrix = factory.eye(num_features).mul(self.lambda_X)
        lambda_IY: Matrix = factory.eye(num_targets).mul(self.lambda_Y)

        # Get covariance matrices
        Cxx: Matrix = X.transpose().mul(X).add(lambda_IX)
        Cyy: Matrix = Y.transpose().mul(Y).add(lambda_IY)
        Cxy: Matrix = X.transpose().mul(Y)

        # Apply A^(-1/2)
        Cxx_inv_sqrt: Matrix = self.pow_minus_half(Cxx)
        Cyy_inv_sqrt: Matrix = self.pow_minus_half(Cyy)

        # Calculate omega for SVD
        omega: Matrix = Cxx_inv_sqrt.mul(Cxy).mul(Cyy_inv_sqrt)

        U: Matrix = omega.svd_U().normalized(0)
        V: Matrix = omega.svd_V().normalized(0)

        C: Matrix = U.get_sub_matrix((0, U.num_rows()), (0, self.kcca))
        D: Matrix = V.get_sub_matrix((0, V.num_rows()), (0, self.kcca))

        self.proj_X = Cxx_inv_sqrt.mul(C)
        self.proj_Y = Cyy_inv_sqrt.mul(D)

        return None

    def pow_minus_half(self, A: Matrix) -> Matrix:
        """
        Compute A^(-1/2) = (A^(-1))^(1/2) on a matrix A, where A^(1/2) = M
        with A = MM.

        :param A:   Input matrix.
        :return:    A^(-1/2).
        """
        eig_vals_desc: Matrix = A.get_eigenvalues_sorted_descending()
        eig_vecs_desc: Matrix = A.get_eigenvectors(True)
        diag: Matrix = factory.diag(eig_vals_desc)
        d_sqrt_inv: Matrix = diag.sqrt().inverse()
        A_pow_half: Matrix = eig_vecs_desc.mul(d_sqrt_inv).mul(eig_vecs_desc.transpose())
        return A_pow_half

    def check(self, x: Optional[Matrix], y: Optional[Matrix]) -> Optional[str]:
        """
        Hook method for checking the data before training.

        :param x:   First sample set.
        :param y:   Second sample set.
        :return:    None if successful, otherwise error message.
        """
        if x is None:
            return 'No x matrix provided!'
        if y is None:
            return 'No y matrix provided!'
        return None

    def do_transform(self, predictors: Matrix) -> Matrix:
        """
        Transforms the predictors data.

        :param predictors:  The input data.
        :return:            The transformed data.
        """
        predictors = self.center_X.transform(predictors)
        return predictors.mul(self.proj_X)

    def do_transform_response(self, response: Matrix) -> Matrix:
        """
        Transforms the response data.

        :param response:    The input data.
        :return:            The transformed data.
        """
        response = self.center_Y.transform(response)
        return response.mul(self.proj_Y)

    def transform(self, predictors: Matrix) -> Matrix:
        if not self.is_initialised():
            raise RuntimeError('Algorithm has not been initialised!')

        return self.do_transform(predictors)

    def transform_response(self, response: Matrix) -> Matrix:
        if not self.is_initialised():
            raise RuntimeError('Algorithm has not been initialised!')

        return self.do_transform_response(response)





#  _AbstractSingleResponsePLS.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from abc import ABC
from typing import Optional

from ._AbstractPLS import AbstractPLS
from ...core import real, NAN, PreprocessingType, ONE, ZERO
from ...core.matrix import Matrix, helper
from ...transformation import AbstractTransformation, Center, Standardize


class AbstractSingleResponsePLS(AbstractPLS, ABC):
    def __init__(self):
        super().__init__()
        self.class_mean: real = NAN  # The class mean
        self.class_std_dev: real = NAN  # The class stddev
        self.trans_predictors: Optional[AbstractTransformation] = None  # The transformation for the predictors
        self.trans_response: Optional[AbstractTransformation] = None  # The transformation for the response

    def reset(self):
        """
        Resets the member variables.
        """
        super().reset()

        self.class_mean = NAN
        self.class_std_dev = NAN
        self.trans_predictors = None
        self.trans_response = None

    def check(self, predictors: Optional[Matrix], response: Optional[Matrix]) -> Optional[str]:
        """
        Hook method for checking the data before training.

        :param predictors:  The input data.
        :param response:    The dependent variable(s).
        :return:            None if successful, otherwise error message.
        """
        result = super().check(predictors, response)

        if result is None:
            if response.num_columns() != 1:
                result = 'Algorithm requires exactly one response variable, found: ' + str(response.num_columns())

        return result

    def do_initialize(self, predictors: Matrix, response: Matrix) -> Optional[str]:
        if self.preprocessing_type is PreprocessingType.CENTER:
            self.class_mean = helper.mean(response, 0)
            self.class_std_dev = ONE
            self.trans_predictors = Center()
            self.trans_response = Center()
        elif self.preprocessing_type is PreprocessingType.STANDARDIZE:
            self.class_mean = helper.mean(response, 0)
            self.class_std_dev = helper.stdev(response, 0)
            self.trans_predictors = Standardize()
            self.trans_response = Standardize()
        elif self.preprocessing_type is PreprocessingType.NONE:
            self.class_mean = ZERO
            self.class_std_dev = ONE
            self.trans_predictors = None
            self.trans_response = None
        else:
            raise RuntimeError('Unhandled preprocessing type: ' + str(self.preprocessing_type))

        if self.trans_predictors is not None:
            self.trans_predictors.configure(predictors)
            predictors = self.trans_predictors.transform(predictors)
        if self.trans_response is not None:
            self.trans_response.configure(response)
            response = self.trans_response.transform(response)

        result = self.do_perform_initialization(predictors, response)

        return result

    def do_predict(self, predictors: Matrix) -> Matrix:
        """
        Performs predictions on the data.

        :param predictors:  The input data.
        :return:            The transformed data and the predictions.
        """
        result = self.do_perform_predictions(predictors)
        if self.trans_response is not None:
            for i in range(result.num_rows()):
                result.set(i, 0, result.get(i, 0) * self.class_std_dev + self.class_mean)

        return result



#  _AbstractPLSTest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
from abc import ABC
from typing import Optional, Tuple

from wai.test.decorators import Test
from wai.ma.algorithm.pls import AbstractPLS
from wai.ma.core.matrix import Matrix

from ...test import AbstractMatrixAlgorithmTest, TestDataset, Tags


class AbstractPLSTest(AbstractMatrixAlgorithmTest, ABC):
    @Test
    def check_transformed_num_components(self, subject: AbstractPLS, bolts: Matrix, bolts_response: Matrix):
        for i in range(1, 5):
            subject.num_components = i
            subject.initialize(bolts, bolts_response)
            transform: Matrix = subject.transform(bolts)
            self.assertEqual(i, transform.num_columns())

            # Reset
            subject = self.instantiate_subject()

    @staticmethod
    def merge_if_not_empty(tag_1: str, tag_2: str) -> str:
        if tag_1 != '' and tag_2 != '':
            return os.sep.join((tag_1, tag_2))
        else:
            return tag_1 + tag_2

    def standard_regression(self, subject: AbstractPLS, *resources: Matrix):
        # Extract data
        bolts, bolts_response = resources

        # Initialise PLS
        results: Optional[str] = subject.initialize(bolts, bolts_response)
        if results is not None:
            self.fail('Algorithm#initialize failed with result: ' + results)

        return self.add_default_pls_matrices(subject, bolts)

    def add_default_pls_matrices(self, algorithm: AbstractPLS, x: Matrix, sub_tag: str = ''):
        """
        Adds default PLS matrices, that is predictions, transformations, loadings and
        model parameter matrices.
        """
        result = {}

        result.update(self.add_predictions(algorithm, x, sub_tag))
        result.update(self.add_transformation(algorithm, x, sub_tag))
        result.update(self.add_loadings(algorithm, sub_tag))
        result.update(self.add_matrices(algorithm, sub_tag))

        return result

    def add_transformation(self, algorithm: AbstractPLS, x: Matrix, sub_tag: str = ''):
        """
        Add transformation to the regression group.
        """
        return {self.merge_if_not_empty(sub_tag, Tags.TRANSFORM): algorithm.transform(x)}

    def add_predictions(self, algorithm: AbstractPLS, x: Matrix, sub_tag: str = ''):
        """
        Add predictions to the regression group.
        """
        # Add predictions
        if algorithm.can_predict():
            preds: Matrix = algorithm.predict(x)
            return {self.merge_if_not_empty(sub_tag, Tags.PREDICTIONS): preds}

        return {}

    def add_loadings(self, algorithm: AbstractPLS, sub_tag: str = ''):
        """
        Add loadings to the regression group.
        """
        # Add loadings
        if algorithm.has_loadings():
            return {self.merge_if_not_empty(sub_tag, Tags.LOADINGS): algorithm.get_loadings()}

        return {}

    def add_matrices(self, algorithm: AbstractPLS, sub_tag: str = ''):
        """
        Add model matrices to the regression group.
        """
        # Add matrices
        result = {}
        for matrix_name in algorithm.get_matrix_names():
            tag: str = Tags.MATRIX + '-' + matrix_name
            result.update({self.merge_if_not_empty(sub_tag, tag): algorithm.get_matrix(matrix_name)})

        return result

    @classmethod
    def get_datasets(cls) -> Tuple[TestDataset, TestDataset]:
        return TestDataset.BOLTS, TestDataset.BOLTS_RESPONSE



#  _CCAFilterTest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Tuple

from wai.test.decorators import RegressionTest
from wai.ma.algorithm import CCAFilter
from wai.ma.core.matrix import Matrix

from ..test import AbstractMatrixAlgorithmTest, Tags, TestDataset


class CCAFilterTest(AbstractMatrixAlgorithmTest):
    @classmethod
    def subject_type(cls):
        return CCAFilter

    @RegressionTest
    def lambda_X_10(self, subject: CCAFilter, *resources: Matrix):
        subject.lambda_X = 10
        return self.standard_regression(subject, *resources)

    @RegressionTest
    def lambda_Y_10(self, subject: CCAFilter, *resources: Matrix):
        subject.lambda_Y = 10
        return self.standard_regression(subject, *resources)

    def standard_regression(self, subject: CCAFilter, *resources: Matrix):
        bolts, bolts_response = resources

        subject.initialize(bolts, bolts_response)

        return {
            Tags.TRANSFORM: subject.transform(bolts),
            Tags.PROJECTION + '-X': subject.proj_X,
            Tags.PROJECTION + '-Y': subject.proj_Y
        }

    @classmethod
    def get_datasets(cls) -> Tuple[TestDataset, TestDataset]:
        return TestDataset.BOLTS, TestDataset.BOLTS_RESPONSE




#  _LogCosH.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Tuple, Optional

from ....core import real
from ....core.matrix import Matrix
from ....core.matrix.factory import zeros, filled
from ._NegEntropyApproximationFunction import NegEntropyApproximationFunction


class LogCosH(NegEntropyApproximationFunction):
    """
    LogCosH Negative Entropy Approximation Function.
    """
    def __init__(self, alpha: real = real(1.0)):
        super().__init__()
        self.alpha = alpha

    def apply(self, x: Optional[Matrix]) -> Tuple[Matrix, Matrix]:
        x = x.mul(self.alpha)
        gx: Matrix = x.tanh()
        g_x: Matrix = zeros(gx.num_rows(), 1)
        ones: Matrix = filled(1, gx.num_columns(), 1.0)
        for i in range(gx.num_rows()):
            gxi: Matrix = gx.get_row(i)
            g_xi = ones.sub(gxi.pow_elementwise(2)).mul(self.alpha).mean()
            g_x.set(i, 0, g_xi)
        return gx, g_x



#  _AbstractKernel.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from abc import ABC, abstractmethod
from typing import Optional

from ...core import real
from ...core.matrix import Matrix, factory


class AbstractKernel(ABC):
    """
    Abstract kernel class. Implementations represent kernels that compute a dot product of two given
    vectors in the kernel space (see AbstractKernel.applyVector(Matrix, Matrix)).
    That is: K(x,y) = phi(x)*phi(y)
    """
    @abstractmethod
    def apply_vector(self, x: Matrix, y: Matrix) -> real:
        """
        Compute the dot product of the mapped x and y vectors in the kernel space, that is:
        K(x,y) = phi(x)*phi(y)

        :param x:   First vector.
        :param y:   Second vector.
        :return:    Dot product of the given vectors in the kernel space.
        """
        pass

    def apply_matrix(self, X: Matrix, Y: Optional[Matrix] = None) -> Matrix:
        """
        Create a matrix K that consists of entries K_i,j = K(x_i,y_j) = phi(x_i)*phi(y_j)

        :param X:   First matrix.
        :param Y:   Second matrix.
        :return:    Matrix K with K_i,j = K(x_i,y_j) = phi(x_i)*phi(y_j)
        """
        if Y is None:
            return self.__apply_matrix_single_arg(X)
        else:
            result: Matrix = factory.zeros(X.num_rows(), Y.num_rows())
            for i in range(X.num_rows()):
                for j in range(Y.num_rows()):
                    row_i: Matrix = X.get_row(i)
                    row_j: Matrix = Y.get_row(j)
                    value: real = self.apply_vector(row_i, row_j)
                    result.set(i, j, value)
            return result

    def __apply_matrix_single_arg(self, X: Matrix) -> Matrix:
        """
        Create a matrix K that consists of entries K_i,j = K(x_i,x_j) = phi(x_i)*phi(x_j)

        :param X:   First matrix.
        :return:    Matrix K with K_i,j = K(x_i,x_j) = phi(x_i)*phi(x_j)
        """
        n: int = X.num_rows()
        result: Matrix = factory.zeros(n, n)
        for i in range(n):
            for j in range(i, n):
                row_i: Matrix = X.get_row(i)
                row_j: Matrix = X.get_row(j)
                value: real = self.apply_vector(row_i, row_j)
                result.set(i, j, value)
                result.set(j, i, value)
        return result



#  _EPO.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional

from ...core.matrix import Matrix, factory
from ._GLSW import GLSW


class EPO(GLSW):
    def __init__(self):
        super().__init__()
        self.N: int = 5

    @staticmethod
    def validate_N(value: int) -> bool:
        return value > 0

    def initialize(self, x1: Optional[Matrix] = None, x2: Optional[Matrix] = None) -> str:
        if x1 is None and x2 is None:
            super().initialize()
            self.N = 5
        else:
            return super().initialize(x1, x2)

    def get_weight_matrix(self, C: Matrix) -> Matrix:
        """
        Instead of calculating D from C, create an identity matrix.

        :param C:   Covariance matrix.
        :return:    Identity matrix.
        """
        return factory.eye(self.N)

    def get_eigenvector_matrix(self, C: Matrix) -> Matrix:
        """
        Only return the first N eigenvectors.

        :param C:   Covariance matrix.
        :return:    Matrix with first N eigenvectors.
        """
        sort_dominance: bool = True
        V: Matrix = C.get_eigenvectors(sort_dominance)
        V = V.get_sub_matrix((0, V.num_rows()), (0, min(V.num_columns(), self.N)))
        return V



#  _Matrix.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from numbers import Number
from typing import Optional, Union, List, Tuple, Callable, Set, IO

import numpy as np

from .._Serialisable import Serialisable
from ..error import MatrixAlgorithmsError, InvalidAxisError, InvalidShapeError
from .._types import real


class Matrix(Serialisable):
    def __init__(self, data: Optional[np.ndarray] = None):
        if data is not None:
            if data.ndim > 2:
                raise ValueError('Matrix only handles 2-dimensional data!')

            while data.ndim < 2:
                data = np.array([data])

        self.data = data
        self.eigenvalue_decomposition: Optional[Tuple[Matrix, Matrix]] = None
        self.singular_value_decomposition: Optional[Tuple[Matrix, Matrix, Matrix]] = None
        self.qr_decomposition = None

        # Alias
        self.t = self.transpose

    def get_sub_matrix(self,
                       rows: Union[List[int], Tuple[int, int]],
                       columns: Union[List[int], Tuple[int, int]]
                       ) -> 'Matrix':
        if isinstance(rows, tuple):
            rows = [i for i in range(*rows)]
        if isinstance(columns, tuple):
            columns = [i for i in range(*columns)]

        # Use advanced indexing
        rows = np.array(rows, dtype=int)
        columns = np.array(columns, dtype=int)
        rows = rows[:, np.newaxis]

        return Matrix(np.array(self.data[rows, columns]))

    def get_rows(self,
                 rows: Union[List[int], Tuple[int, int]]
                 ) -> 'Matrix':
        return self.get_sub_matrix(rows, (0, self.num_columns()))

    def get_columns(self,
                    columns: Union[List[int], Tuple[int, int]]
                    ) -> 'Matrix':
        return self.get_sub_matrix((0, self.num_rows()), columns)

    def get_eigenvectors(self, sort_dominance: bool = False) -> 'Matrix':
        if sort_dominance:
            return self.get_eigenvectors_sorted_descending()

        self.make_eigenvalue_decomposition()
        return self.eigenvalue_decomposition[1].copy()

    def get_eigenvectors_sorted_ascending(self) -> 'Matrix':
        return self.get_eigenvectors_sorted(True)

    def get_eigenvectors_sorted_descending(self) -> 'Matrix':
        return self.get_eigenvectors_sorted(False)

    def get_dominant_eigenvector(self) -> 'Matrix':
        return self.get_eigenvectors_sorted_descending().get_column(0)

    def get_eigenvalue_decomposition_V(self) -> 'Matrix':
        self.make_eigenvalue_decomposition()
        return self.eigenvalue_decomposition[1]

    def get_eigenvalue_decomposition_D(self) -> 'Matrix':
        self.make_eigenvalue_decomposition()
        return self.eigenvalue_decomposition[0].vector_to_diagonal()

    def get_eigenvalues(self) -> 'Matrix':
        self.make_eigenvalue_decomposition()
        return self.eigenvalue_decomposition[0].copy()

    def get_eigenvalues_sorted_descending(self) -> 'Matrix':
        return self.get_eigenvalues_sorted(False)

    def get_eigenvalues_sorted_ascending(self) -> 'Matrix':
        return self.get_eigenvalues_sorted(True)

    def make_singular_value_decomposition(self):
        u, s, vh = np.linalg.svd(self.data, full_matrices=False)
        # u and vh matrices are multiplied by -1 to match OJAlgo reference implementation
        self.singular_value_decomposition = (Matrix(u).mul(-1), Matrix(s), Matrix(vh).mul(-1))

    def make_eigenvalue_decomposition(self):
        """
        Initialises the eigenvalue decomposition.

        N.B. The numpy eigenvalue solver may produce different eigenvalues/vectors to
        the reference solver (OJAlgo in the Java matrix-algorithms repo). While these
        are still correct solutions to the decomposition, implementations that rely on
        the exact direction of the eigenvectors may be affected. Possible future solution
        is to enforce normalisation and a standard quadrant that the vectors must point into.
        """
        if self.eigenvalue_decomposition is None:
            eigenvalues, eigenvectors = np.linalg.eig(self.data)
            self.eigenvalue_decomposition = (Matrix(eigenvalues).transpose(), Matrix(eigenvectors))

    def make_qr_decomposition(self):
        # TODO
        raise NotImplementedError

    def svd_U(self) -> 'Matrix':
        self.make_singular_value_decomposition()
        return self.singular_value_decomposition[0].copy()

    def svd_V(self) -> 'Matrix':
        self.make_singular_value_decomposition()
        return self.singular_value_decomposition[2].copy()

    def svd_S(self) -> 'Matrix':
        self.make_singular_value_decomposition()
        n = self.singular_value_decomposition[1].num_elements()
        raw = [[self.singular_value_decomposition[1].get_from_vector(i) if i == j else 0
                for i in range(n)]
               for j in range(n)]
        return Matrix(np.array(raw, dtype=real))

    def get_singular_values(self) -> 'Matrix':
        return self.singular_value_decomposition[1].transpose()

    def sum(self, axis: Optional[int] = None) -> Union['Matrix', real]:
        if axis is None or axis == -1:
            result = np.add.reduce(np.add.reduce(self.data))
            if axis == -1:
                result = Matrix(result)
            return result
        elif axis == 0 or axis == 1:
            result = Matrix(np.add.reduce(self.data, axis=axis))
            if axis == 1:
                result = result.transpose()
            return result
        else:
            raise ValueError('Acceptable values for axis are None, -1, 0 or 1. ' + str(axis) + 'provided.')

    def norm1(self) -> real:
        return real(np.linalg.norm(np.linalg.norm(self.data, 1, 0), 1))

    def norm2(self) -> real:
        frobenius_norm: real = real(np.linalg.norm(self.data))
        if self.is_vector():
            return frobenius_norm
        else:
            return frobenius_norm / np.ma.sqrt(min(self.num_rows(), self.num_columns()))

    def norm2_squared(self) -> real:
        norm2 = self.norm2()
        return np.multiply(norm2, norm2)

    def mul(self, other: Union['Matrix', Number]) -> 'Matrix':
        if isinstance(other, Matrix):
            if not self.is_multiplicable_with(other):
                raise ValueError('Matrix shapes not multiplicable!')
            else:
                return Matrix(np.matmul(self.data, other.data))
        else:
            return Matrix(np.multiply(self.data, other))

    def is_multiplicable_with(self, other: 'Matrix') -> bool:
        return self.num_columns() == other.num_rows()

    def vector_dot(self, other: 'Matrix') -> real:
        if not self.is_vector() or not other.is_vector():
            raise ValueError('Both matrices must be vectors to perform vector dot operation')

        if self.data.size != other.data.size:
            raise ValueError('Both matrices must be the same length')

        if self.is_column_vector():
            a = self.data.transpose()[0]
        else:
            a = self.data[0]

        if other.is_column_vector():
            b = other.data.transpose()[0]
        else:
            b = other.data[0]

        return np.dot(a, b)

    def normalized(self, axis: int = 0) -> 'Matrix':
        norms = np.linalg.norm(self.data, 2, axis, keepdims=True)
        result = np.divide(self.data, norms)
        return Matrix(result)

    def is_vector(self) -> bool:
        return self.data.shape[0] == 1 or self.data.shape[1] == 1

    def same_shape_as(self, other: 'Matrix') -> bool:
        return self.num_rows() == other.num_rows() and self.num_columns() == other.num_columns()

    def mul_elementwise(self, other: 'Matrix') -> 'Matrix':
        if not self.same_shape_as(other):
            throw_invalid_shapes(self, other)
        return Matrix(np.multiply(self.data, other.data))

    def scale_by_row_vector(self, vector: 'Matrix') -> 'Matrix':
        must_be_column_vector(vector)
        dimensions_must_match(self, vector, columns_to_rows=True)
        return self.mul_by_vector(vector.transpose())

    def scale_by_column_vector(self, vector: 'Matrix') -> 'Matrix':
        must_be_column_vector(vector)
        dimensions_must_match(self, vector, rows_to_rows=True)
        return self.mul_by_vector(vector)

    def add_by_vector(self, vector: 'Matrix') -> 'Matrix':
        return self.vector_op(vector, np.add)

    def div_elementwise(self, other: 'Matrix') -> 'Matrix':
        if not self.same_shape_as(other):
            throw_invalid_shapes(self, other)
        return Matrix(np.divide(self.data, other.data))

    def div(self, scalar: Number) -> 'Matrix':
        return Matrix(np.divide(self.data, scalar))

    def sub(self, other: Union['Matrix', Number]) -> 'Matrix':
        if isinstance(other, Matrix):
            must_be_same_shape(self, other)
            other = other.data
        return Matrix(np.subtract(self.data, other))

    def add(self, other: Union['Matrix', Number]) -> 'Matrix':
        if isinstance(other, Matrix):
            must_be_same_shape(self, other)
            other = other.data
        return Matrix(np.add(self.data, other))

    def pow_elementwise(self, exponent: Number) -> 'Matrix':
        return Matrix(np.power(self.data, exponent))

    def sqrt(self) -> 'Matrix':
        return Matrix(np.sqrt(self.data))

    def transpose(self) -> 'Matrix':
        return Matrix(np.array(self.data.transpose(), copy=True))

    def num_columns(self) -> int:
        return self.data.shape[1]

    def num_rows(self) -> int:
        return self.data.shape[0]

    def get(self, row: int, column: int) -> real:
        return self.data[row][column]

    def set(self, row: int, column: int, value: Number):
        self.reset_cache()
        self.data[row][column] = real(value)

    def set_row(self, row_index: int, row: 'Matrix'):
        must_be_vector(row)
        if row.is_column_vector():
            row = row.transpose()
        replacement_length = min(row.num_columns(), self.num_columns())
        self.reset_cache()
        self.data[row_index, :replacement_length] = row.data[0, :replacement_length]

    def set_column(self, column_index: int, column: 'Matrix'):
        must_be_vector(column)
        if column.is_row_vector():
            column = column.transpose()
        replacement_length = min(column.num_rows(), self.num_rows())
        self.reset_cache()
        self.data[:replacement_length, column_index] = column.data[:replacement_length, 0]

    def get_row(self, row_index: int) -> 'Matrix':
        return Matrix(np.array(self.data[row_index]))

    def get_column(self, column_index: int) -> 'Matrix':
        return Matrix(np.array(self.data[:, column_index])).transpose()

    def inverse(self) -> 'Matrix':
        if self.is_square():
            return Matrix(np.linalg.inv(self.data))
        else:
            raw = [[1 if i == j else 0 for i in range(self.num_rows())] for j in range(self.num_rows())]
            rhs = np.array(raw, dtype=real)
            return Matrix(np.linalg.lstsq(self.data, rhs, rcond=None)[0])

    def copy(self) -> 'Matrix':
        return Matrix(np.copy(self.data))

    def as_real(self) -> real:
        if self.num_rows() != 1 or self.num_columns() != 1:
            raise MatrixAlgorithmsError('Method Matrix#as_real is invalid ' +
                                        'when number of rows != 1 or number of columns != 1.')
        return self.get(0, 0)

    def to_raw_copy_1D(self) -> List[real]:
        copy = []

        for col in range(self.num_columns()):
            for row in range(self.num_rows()):
                copy.append(self.get(row, col))

        return copy

    def to_raw_copy_2D(self) -> List[List[real]]:
        copy = []

        for row in range(self.num_rows()):
            r = []
            for col in range(self.num_columns()):
                r.append(self.get(row, col))
            copy.append(r)

        return copy

    def concat(self, other: 'Matrix', axis: int) -> 'Matrix':
        return Matrix(np.concatenate((self.data, other.data), axis))

    def concat_along_rows(self, other: 'Matrix') -> 'Matrix':
        return self.concat(other, 0)

    def concat_along_columns(self, other: 'Matrix') -> 'Matrix':
        return self.concat(other, 1)

    def reset_cache(self):
        self.eigenvalue_decomposition = None
        self.singular_value_decomposition = None
        self.qr_decomposition = None

    def is_row_vector(self) -> bool:
        return self.is_vector() and self.num_rows() == 1

    def is_column_vector(self) -> bool:
        return self.is_vector() and self.num_columns() == 1

    def apply_elementwise(self, body: Callable[[real], Number]) -> 'Matrix':
        result = self.copy()
        for row in range(self.num_rows()):
            for col in range(self.num_columns()):
                result.data[row][col] = real(body(result.data[row][col]))
        return result

    def clip(self, lower_bound: Number, upper_bound: Number) -> 'Matrix':
        if lower_bound > upper_bound:
            raise MatrixAlgorithmsError('Invalid clipping values. Lower ' +
                                        'bound must be below upper bound')

        lower_bound = real(lower_bound)
        upper_bound = real(upper_bound)

        def clip(el: real) -> real:
            if el < lower_bound:
                return lower_bound
            elif el > upper_bound:
                return upper_bound
            else:
                return el

        return self.apply_elementwise(clip)

    def clip_lower(self, lower_bound: Number) -> 'Matrix':
        return self.clip(lower_bound, np.inf)

    def clip_upper(self, upper_bound: Number) -> 'Matrix':
        return self.clip(-np.inf, upper_bound)

    def sign(self) -> 'Matrix':
        def sign(el: real) -> real:
            if el < 0:
                return real(-1)
            elif el > 0:
                return real(1)
            else:
                return real(0)

        return self.apply_elementwise(sign)

    def abs(self) -> 'Matrix':
        def abs(el: real) -> real:
            if el < 0:
                return -el
            else:
                return el

        return self.apply_elementwise(abs)

    def max(self) -> real:
        return np.amax(self.data)

    def median(self) -> real:
        return np.median(self.data)

    def where_vector(self, condition: Callable[[real], bool]) -> List[int]:
        must_be_vector(self)
        if self.is_row_vector():
            return [i for i in range(self.num_columns())
                    if condition(self.get(0, i))]
        else:
            return [i for i in range(self.num_rows())
                    if condition(self.get(i, 0))]

    def head(self, n: int = 5) -> 'Matrix':
        return Matrix(np.array(self.data[:n, :]))

    def diag(self) -> 'Matrix':
        raw = [[self.get(i, i)] for i in range(min(self.num_rows(), self.num_columns()))]
        return Matrix(np.array(raw))

    def mean(self, axis: Optional[int] = None) -> Union['Matrix', real]:
        if axis is None or axis == -1:
            mean = np.mean(self.data)
            if axis == -1:
                mean = Matrix(mean)
            return mean
        elif axis == 0:
            return Matrix(np.mean(self.data, axis))
        elif axis == 1:
            return Matrix(np.mean(self.data, axis)).transpose()
        else:
            raise InvalidAxisError(axis)

    def reduce_rows_L1(self) -> 'Matrix':
        # TODO
        raise NotImplementedError

    def reduce_columns_L1(self) -> 'Matrix':
        # TODO
        raise NotImplementedError

    def reduce_rows_L2(self) -> 'Matrix':
        # TODO
        raise NotImplementedError

    def reduce_columns_L2(self) -> 'Matrix':
        # TODO
        raise NotImplementedError

    def qr_Q(self) -> 'Matrix':
        # TODO
        raise NotImplementedError

    def qr_R(self) -> 'Matrix':
        # TODO
        raise NotImplementedError

    def contains_NaN(self) -> bool:
        NaN = real('NaN')
        def element_is_NaN(el):
            return el == NaN
        return self.any(element_is_NaN)

    def any(self, function: Callable[[real], bool]) -> bool:
        for row in range(self.num_rows()):
            for col in range(self.num_columns()):
                if function(self.data[row][col]):
                    return True
        return False

    def all(self, function: Callable[[real], bool]) -> bool:
        return not self.any(lambda el: not function(el))

    def which(self, function: Callable[[real], bool]) -> List[Tuple[int, int]]:
        result: List[Tuple[int, int]] = []
        for row in range(self.num_rows()):
            for col in range(self.num_columns()):
                if function(self.data[row][col]):
                    result.append((row, col))
        return result

    def trace(self) -> real:
        return self.diag().sum()

    def __str__(self):
        string = self.row_str(0)
        for row_index in range(1, self.num_rows()):
            string += '\n' + self.row_str(row_index)
        return string

    def row_str(self, row_index: int) -> str:
        row = self.data[row_index]
        row_string = str(row[0])
        for column_index in range(1, self.num_columns()):
            row_string += ',' + str(row[column_index])
        return row_string

    def __eq__(self, other):
        if self is other:
            return True

        if other is None or self.__class__ is not other.__class__:
            return False

        diff: Matrix = self.sub(other)

        return not diff.data.any()

    def __hash__(self):
        # TODO
        raise NotImplementedError

    def shape_string(self) -> str:
        return '[' +\
               str(self.num_rows()) +\
               ' x ' +\
               str(self.num_columns()) +\
               ']'

    def sub_by_vector(self, vector: 'Matrix') -> 'Matrix':
        return self.vector_op(vector, np.subtract)

    def mul_by_vector(self, vector: 'Matrix') -> 'Matrix':
        return self.vector_op(vector, np.multiply)

    def add_by_vector_modify(self, vector: 'Matrix'):
        self.vector_op_modify(vector, np.add)

    def sub_by_vector_modify(self, vector: 'Matrix'):
        self.vector_op_modify(vector, np.subtract)

    def mul_by_vector_modify(self, vector: 'Matrix'):
        self.vector_op_modify(vector, np.multiply)

    def div_by_vector_modify(self, vector: 'Matrix'):
        self.vector_op_modify(vector, np.divide)

    def vector_op(self, vector: 'Matrix', op: np.ufunc) -> 'Matrix':
        result = self.copy()
        result.vector_op_modify(vector, op)
        return result

    def vector_op_modify(self, vector: 'Matrix', op: np.ufunc):
        must_be_vector(vector)
        self.ensure_vector_size_matches(vector)
        op.at(self.data, ..., vector.data)

    def vector_size_matches(self, vector: 'Matrix') -> bool:
        if not vector.is_vector():
            return False
        elif vector.is_row_vector():
            return self.num_columns() == vector.num_columns()
        else:
            return self.num_rows() == vector.num_rows()

    def ensure_vector_size_matches(self, vector: 'Matrix'):
        if not self.vector_size_matches(vector):
            raise ValueError('The supplied vector is not the right size')

    def num_elements(self):
        return self.num_rows() * self.num_columns()

    def get_from_vector(self, index: int):
        must_be_vector(self)
        if self.is_row_vector():
            return self.get(0, index)
        else:
            return self.get(index, 0)

    def vector_to_diagonal(self):
        must_be_vector(self)
        raw = [[self.get_from_vector(i) if i == j else real(0) for i in range(self.num_elements())]
               for j in range(self.num_elements())]
        return Matrix(np.array(raw))

    def get_eigenvectors_sorted(self, ascending: bool):
        self.make_eigenvalue_decomposition()

        # Get eigenpairs
        eigenpairs = [(self.eigenvalue_decomposition[0].get_from_vector(i),
                       self.eigenvalue_decomposition[1].get_column(i))
                      for i in range(self.num_rows())]

        eigenpairs.sort(key=lambda v: v[0], reverse=not ascending)

        first = eigenpairs[0][1]

        for i in range(1, len(eigenpairs)):
            first = first.concat_along_columns(eigenpairs[i][1])

        return first

    def get_eigenvalues_sorted(self, ascending: bool) -> 'Matrix':
        eigenvalues = self.get_eigenvalues()
        eigenvalues.data.sort(0)
        if not ascending:
            eigenvalues.data = np.flip(eigenvalues.data)
        return eigenvalues

    def exp(self):
        return Matrix(np.exp(self.data))

    def tanh(self):
        return Matrix(np.tanh(self.data))

    def is_square(self):
        return self.num_rows() == self.num_columns()

    def pseudo_inverse(self):
        return Matrix(np.linalg.pinv(self.data))

    def serialise_state(self, stream: IO[bytes]):
        # Write the number of rows and columns
        stream.write(self.serialise_to_bytes(self.num_rows()))
        stream.write(self.serialise_to_bytes(self.num_columns()))

        # Write each data element in order
        for row_index in range(self.num_rows()):
            for column_index in range(self.num_columns()):
                stream.write(self.serialise_to_bytes(self.get(row_index, column_index)))


def must_be_row_vector(vector: 'Matrix'):
    if not vector.is_row_vector():
        raise ValueError('Must be a row vector')


def must_be_column_vector(vector: 'Matrix'):
    if not vector.is_column_vector():
        raise ValueError('Must be a column vector')


def must_be_vector(vector: 'Matrix'):
    if not vector.is_vector():
        raise ValueError('Must be a vector')


def must_be_same_shape(m1: Matrix, m2: Matrix):
    dimensions_must_match(m1, m2, rows_to_rows=True, columns_to_columns=True)


def dimensions_must_match(m1: Matrix, m2: Matrix, *,
                          rows_to_rows: bool = False,
                          rows_to_columns: bool = False,
                          columns_to_rows: bool = False,
                          columns_to_columns: bool = False):
    checks: Set[bool] = {not rows_to_rows or m1.num_rows() == m2.num_rows(),
                         not rows_to_columns or m1.num_rows() == m2.num_columns(),
                         not columns_to_rows or m1.num_columns() == m2.num_rows(),
                         not columns_to_columns or m1.num_columns() == m2.num_columns()}

    if False in checks:
        raise InvalidShapeError('', m1.shape_string(), m2.shape_string())


def throw_invalid_shapes(m1: Matrix, m2: Matrix):
    raise InvalidShapeError('Invalid matrix multiplication. Shapes ' +
                            m1.shape_string() +
                            ' and ' +
                            m2.shape_string() +
                            ' do not match.')



#  _KernelPLS.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional, List

from ...transformation.kernel import AbstractKernel, RBFKernel
from ._AbstractMultiResponsePLS import AbstractMultiResponsePLS
from ...core import real, ZERO, ONE
from ...core.matrix import Matrix, factory
from ...transformation import Center


class KernelPLS(AbstractMultiResponsePLS):
    """
    Kernel Partial Least Squares algorithm.

    See here:
    <a href="http://www.jmlr.org/papers/volume2/rosipal01a/rosipal01a.pdf">
        Kernel Partial Least Squares Regression in Reproducing Kernel Hilbert Space
    </a>
    """
    SEED: int = 0

    def __init__(self):
        super().__init__()
        self.K_orig: Optional[Matrix] = None  # Calibration data in feature space
        self.K_deflated: Optional[Matrix] = None
        self.T: Optional[Matrix] = None  # Scores on K
        self.U: Optional[Matrix] = None  # Scores on Y
        self.P: Optional[Matrix] = None  # Loadings on K
        self.Q: Optional[Matrix] = None  # Loadings on Y
        self.B_RHS: Optional[Matrix] = None  # Partial regression matrix
        self.X: Optional[Matrix] = None  # Training points
        self.kernel: Optional[AbstractKernel] = None  # Kernel for feature transformation
        self.tol: real = real(1e-6)  # Inner NIPALS loop improvement tolerance
        self.max_iter: int = 500  # Inner NIPALS loop maximum number of iterations
        self.center_X: Optional[Center] = None  # Center X transformation
        self.center_Y: Optional[Center] = None  # Center Y transformation

    def initialize(self, predictors: Optional[Matrix] = None, response: Optional[Matrix] = None) -> Optional[str]:
        if predictors is None and response is None:
            super().initialize()
            self.kernel = RBFKernel()
            self.tol = real(1e-6)
            self.max_iter = 500
            self.center_X = Center()
            self.center_Y = Center()
        else:
            return super().initialize(predictors, response)

    @staticmethod
    def validate_kernel(value: AbstractKernel) -> bool:
        return True

    @staticmethod
    def validate_max_iter(value: int) -> bool:
        return value >= 0

    @staticmethod
    def validate_tol(value: real) -> bool:
        return value >= ZERO

    def get_min_columns_response(self) -> int:
        return 1

    def get_max_columns_response(self) -> int:
        return -1

    def do_perform_initialization(self, predictors: Matrix, response: Matrix) -> Optional[str]:
        # Init
        num_components: int = self.num_components
        self.X = predictors
        self.X = self.center_X.transform(self.X)
        Y: Matrix = response
        Y = self.center_Y.transform(Y)

        num_rows: int = self.X.num_rows()
        num_classes: int = Y.num_columns()

        q: Matrix = factory.zeros(num_classes, 1)
        t: Matrix = factory.zeros(num_rows, 1)
        w: Matrix = factory.zeros(num_rows, 1)
        I: Matrix = factory.eye(num_rows, num_rows)

        self.T = factory.zeros(num_rows, num_components)
        self.U = factory.zeros(num_rows, num_components)
        self.P = factory.zeros(num_rows, num_components)
        self.Q = factory.zeros(num_classes, num_components)

        self.K_orig = self.kernel.apply_matrix(self.X)
        self.K_orig = self.centralize_train_in_kernel_space(self.K_orig)
        self.K_deflated = self.K_orig.copy()

        for current_component in range(num_components):
            iterations: int = 0
            u_old: Optional[Matrix] = None
            u: Matrix = factory.randn(num_rows, 1, KernelPLS.SEED + current_component)
            iteration_change: real = real(self.tol * 10)

            # Repeat 1) - 3) until convergence: either change of u is lower than self.tol or maximum
            # number of iterations has been reached (self.max_iter)
            while iteration_change > self.tol and iterations < self.max_iter:
                # 1)
                t: Matrix = self.K_deflated.mul(u).normalized()
                w: Matrix = t.copy()

                # 2)
                q = Y.transpose().mul(t)

                # 3)
                u_old = u
                u = Y.mul(q).normalized()

                # Update stopping conditions
                iterations += 1
                iteration_change = u.sub(u_old).norm2()

            # Deflate
            t_t_trans: Matrix = t.mul(t.transpose())
            part: Matrix = I.sub(t_t_trans)

            self.K_deflated = part.mul(self.K_deflated).mul(part)
            Y = Y.sub(t.mul(q.transpose()))
            p: Matrix = self.K_deflated.transpose().mul(w).div(w.transpose().mul(w).as_real())

            # Store u,t,q,p
            self.T.set_column(current_component, t)
            self.U.set_column(current_component, u)
            self.Q.set_column(current_component, q)
            self.P.set_column(current_component, p)

        # Calculate right hand side of the regression matrix B
        tT_times_K_times_U = self.T.transpose().mul(self.K_orig).mul(self.U)
        inv: Matrix = tT_times_K_times_U.inverse()
        self.B_RHS = inv.mul(self.Q.transpose())
        return None

    def centralize_train_in_kernel_space(self, K: Matrix) -> Matrix:
        """
        Centralize a kernel matrix in the kernel space via:
        K <- (I - 1/n * 1_n * 1_n^T) * K * (I - 1/n * 1_n * 1_n^T)

        :param K:   Kernel matrix.
        :return:    Centralised kernel matrix.
        """
        n: int = self.X.num_rows()
        I: Matrix = factory.eye(n, n)
        one: Matrix = factory.filled(n, 1, ONE)

        # Centralize in kernel space
        part: Matrix = I.sub(one.mul(one.transpose()).div(n))
        return part.mul(K).mul(part)

    def centralize_test_in_kernel_space(self, K: Matrix) -> Matrix:
        """
        :param K:   Kernel matrix.
        :return:    Centralised kernel matrix.
        """
        n_train: int = self.X.num_rows()
        n_test: int = K.num_rows()
        I: Matrix = factory.eye(n_train, n_train)
        ones_train_test_scaled: Matrix = factory.filled(n_test, n_train, real(ONE / n_train))

        ones_train_scaled = factory.filled(n_train, n_train, real(ONE / n_train))
        return (K.sub(ones_train_test_scaled.mul(self.K_orig))).mul(I.sub(ones_train_scaled))

    def do_perform_predictions(self, predictors: Matrix) -> Matrix:
        K_t: Matrix = self.do_transform(predictors)
        Y_hat: Matrix = K_t.mul(self.B_RHS)
        Y_hat = self.center_Y.inverse_transform(Y_hat)
        return Y_hat

    def do_transform(self, predictors: Matrix) -> Matrix:
        predictors_centered = self.center_X.transform(predictors)
        K_t: Matrix = self.kernel.apply_matrix(predictors_centered, self.X)
        K_t = self.centralize_test_in_kernel_space(K_t)

        return K_t.mul(self.U)

    def get_matrix_names(self) -> List[str]:
        return ['K', 'T', 'U', 'P', 'Q']

    def get_matrix(self, name: str) -> Optional[Matrix]:
        if name == 'K':
            return self.K_deflated
        elif name == 'T':
            return self.T
        elif name == 'U':
            return self.U
        elif name == 'P':
            return self.P
        elif name == 'Q':
            return self.Q
        return None

    def has_loadings(self) -> bool:
        return True

    def reset(self):
        super().reset()
        self.K_orig = None
        self.K_deflated = None
        self.T = None
        self.U = None
        self.P = None
        self.Q = None
        self.B_RHS = None
        self.X = None
        self.center_X = Center()
        self.center_Y = Center()

    def get_loadings(self) -> Optional[Matrix]:
        return self.T

    def can_predict(self) -> bool:
        return True



#  _OPLSTest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from wai.test.decorators import Test, Skip, RegressionTest

from wai.ma.algorithm.pls import OPLS, NIPALS, KernelPLS, PLS1, SIMPLS, SparsePLS, PRM, CCARegression
from wai.ma.core.matrix import Matrix

from ._AbstractPLSTest import AbstractPLSTest


class OPLSTest(AbstractPLSTest):
    @classmethod
    def subject_type(cls):
        return OPLS

    @Test
    @Skip("OPLS.transform(X) removes the signal from X_test that is"
          "orthogonal to y_train and does not change its shape")
    def check_transformed_num_components(self):
        # Do nothing since OPLS.transform(X) removes the signal from X_test that is
        # orthogonal to y_train and does not change its shape
        super().check_transformed_num_components()

    @RegressionTest
    def base_NIPALS(self, subject: OPLS, *resources: Matrix):
        subject.base_PLS = NIPALS()
        return self.standard_regression(subject, *resources)

    @RegressionTest
    def base_KernelPLS(self, subject: OPLS, *resources: Matrix):
        subject.base_PLS = KernelPLS()
        return self.standard_regression(subject, *resources)

    @RegressionTest
    def base_PLS1(self, subject: OPLS, *resources: Matrix):
        subject.base_PLS = PLS1()
        return self.standard_regression(subject, *resources)

    @RegressionTest
    def base_SIMPLS(self, subject: OPLS, *resources: Matrix):
        subject.base_PLS = SIMPLS()
        return self.standard_regression(subject, *resources)

    @RegressionTest
    def base_CCARegression(self, subject: OPLS, *resources: Matrix):
        subject.base_PLS = CCARegression()
        return self.standard_regression(subject, *resources)

    @RegressionTest
    def base_PRM(self, subject: OPLS, *resources: Matrix):
        subject.base_PLS = PRM()
        return self.standard_regression(subject, *resources)

    @RegressionTest
    def base_SparsePLS(self, subject: OPLS, *resources: Matrix):
        subject.base_PLS = SparsePLS()
        return self.standard_regression(subject, *resources)



#  _RobustScaler.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional, Tuple

from sklearn.preprocessing import RobustScaler as Intern

from ..core.matrix import Matrix
from ._AbstractTransformation import AbstractTransformation


class RobustScaler(AbstractTransformation):
    # The output distribution options
    OUTPUT_DISTRIBUTIONS = ("uniform", "normal")

    def __init__(self):
        super().__init__()

        self.with_centering: bool = True
        self.with_scaling: bool = True
        self.quantile_range: Tuple[float, float] = (25.0, 75.0)

        self._intern: Optional[Intern] = None

    def __setattr__(self, key, value):
        # Validate values
        validator_name = 'validate_' + key
        if hasattr(self, validator_name):
            validator = getattr(self, validator_name)
            validator(value)

            # If validation passes, reset
            self.reset()

        super().__setattr__(key, value)

    def validate_quantile_range(self, selection: Tuple[float, float]):
        if not (0.0 < selection[0] < selection[1] < 100.0):
            raise ValueError(f"Quantile range must be 0.0 < q_min < q_max < 100.0. "
                             f"Got q_min={selection[0]}, q_max={selection[1]}")

    def reset(self):
        super().reset()
        self._intern = None

    def configure(self, data: Matrix):
        self._intern = Intern(with_centering=self.with_centering,
                              with_scaling=self.with_scaling,
                              quantile_range=self.quantile_range)
        self._intern.fit(data.data)
        self.configured = True

    def do_transform(self, data: Matrix) -> Matrix:
        return Matrix(self._intern.transform(data.data))

    def do_inverse_transform(self, data: Matrix) -> Matrix:
        return Matrix(self._intern.inverse_transform(data.data))



#  _FastICATest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Tuple

from wai.test.decorators import RegressionTest

from wai.ma.algorithm.ica import FastICA, Algorithm
from wai.ma.algorithm.ica.approxfun import LogCosH, Cube, Exponential
from wai.ma.core.matrix import Matrix

from ...test import AbstractMatrixAlgorithmTest, TestDataset


class FastICATest(AbstractMatrixAlgorithmTest):
    @classmethod
    def subject_type(cls):
        return FastICA

    @RegressionTest
    def deflation(self, subject: FastICA, *resources: Matrix):
        subject.algorithm = Algorithm.DEFLATION
        return self.standard_regression(subject, *resources)

    @RegressionTest
    def parallel(self, subject: FastICA, *resources: Matrix):
        subject.algorithm = Algorithm.PARALLEL
        return self.standard_regression(subject, *resources)

    @RegressionTest
    def white_false(self, subject: FastICA, *resources: Matrix):
        subject.whiten = False
        return self.standard_regression(subject, *resources)

    @RegressionTest
    def logcosh(self, subject: FastICA, *resources: Matrix):
        subject.fun = LogCosH()
        return self.standard_regression(subject, *resources)

    @RegressionTest
    def cube(self, subject: FastICA, *resources: Matrix):
        subject.fun = Cube()
        return self.standard_regression(subject, *resources)

    @RegressionTest
    def exp(self, subject: FastICA, *resources: Matrix):
        subject.fun = Exponential()
        return self.standard_regression(subject, *resources)

    def standard_regression(self, subject: FastICA, *resources: Matrix):
        X, = resources

        transform: Matrix = subject.transform(X)

        return {
            'transform': transform,
            'components': subject.components,
            'mixing': subject.mixing,
            'sources': subject.sources
        }

    @classmethod
    def get_datasets(cls) -> Tuple[TestDataset]:
        return TestDataset.BOLTS,



#  _EquidistanceFilter.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from ..core import Filter
from ..core.matrix import Matrix, factory


class Equidistance(Filter):
    """
    Filter which resamples rows to the given number of columns
    using linear interpolation.
    """
    def __init__(self, num_samples: int):
        self.num_samples: int = num_samples  # The number of samples to resample to

    def __setattr__(self, key, value):
        # Must be at least 2 samples in the output
        if key == "num_samples" and value < 2:
            raise ValueError("num_samples must be at least 2")

        super().__setattr__(key, value)

    def transform(self, predictors: Matrix) -> Matrix:
        # Create a 1-dimensional space where the sample positions
        # for the original and transformed matrices are integral
        resampled_step = (predictors.num_columns() - 1)
        original_step = (self.num_samples - 1)

        # Create a correctly-sized matrix to hold the resampled columns
        resampled = factory.zeros(predictors.num_rows(), self.num_samples)

        # Calculate each resampled column in turn
        for resampled_index in range(self.num_samples):
            # Get the x-position of this sample in our integral space
            x = resampled_step * resampled_index

            # Get the index of the original column at (or just
            # to the left of) this x-position
            old_index = x // original_step

            # If there is an original column at exactly this position,
            # just copy it
            if x % original_step == 0:
                resampled.set_column(resampled_index, predictors.get_column(old_index))
                continue

            # Get the columns either side of the x-position
            left = predictors.get_column(old_index)
            right = predictors.get_column(old_index + 1)

            # Get the interpolation factor for the x-position
            t = (x % original_step) / original_step

            # Add the linear interpolation of the 2 columns
            resampled.set_column(resampled_index, left.mul(1 - t).add(right.mul(t)))

        return resampled




#  _RealSerialiser.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import IO, Optional

from wai.ma.core import real
from wai.test.serialisation import RegressionSerialiser

from ._epsilon import EPSILON


class RealSerialiser(RegressionSerialiser[real]):
    @classmethod
    def binary(cls) -> bool:
        return False

    @classmethod
    def extension(cls) -> str:
        return "txt"

    @classmethod
    def serialise(cls, result: real, file: IO[str]):
        file.write(str(result))

    @classmethod
    def deserialise(cls, file: IO[str]) -> real:
        return real(file.read())

    @classmethod
    def compare(cls, result: real, reference: real) -> Optional[str]:
        if abs(result - reference) > EPSILON:
            return "Result: " + str(result) + ", reference: " + str(reference)
        else:
            return None



#  _Center.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional

from ..core.matrix import Matrix
from ..core.matrix.helper import column_means
from ._AbstractTransformation import AbstractTransformation


class Center(AbstractTransformation):
    def __init__(self):
        super().__init__()
        self.means: Optional[Matrix] = None

    def reset(self):
        super().reset()
        self.means = None

    def configure(self, data: Matrix):
        self.means = column_means(data)
        self.configured = True

    def do_transform(self, data: Matrix) -> Matrix:
        result = data.copy()
        result.sub_by_vector_modify(self.means)
        return result

    def do_inverse_transform(self, data: Matrix) -> Matrix:
        result = data.copy()
        result.add_by_vector_modify(self.means)
        return result



#  _CCARegression.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from ._NIPALS import NIPALS, WeightCalculationMode, DeflationMode


class CCARegression(NIPALS):
    def get_weight_calculation_mode(self) -> 'WeightCalculationMode':
        return WeightCalculationMode.CCA  # Mode B in sklearn

    def __getattribute__(self, item):
        if item == 'deflation_mode':
            return DeflationMode.CANONICAL
        else:
            return super().__getattribute__(item)

    def __setattr__(self, key, value):
        if key == 'deflation_mode':
            if value is not DeflationMode.CANONICAL:
                self.logger.warning('CCARegression only allows CANONICAL deflation mode.')
        else:
            super().__setattr__(key, value)
